"""Configuration file for the Sphinx documentation builder."""
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "dns_exporter"
copyright = "2023-%Y, Thomas Steen Rasmussen"
author = "Thomas Steen Rasmussen"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_changelog",
]


templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ["_static"]



"""``dns_exporter.config`` contains all the configuration related code for dns_exporter.

The primary class is the Config object and the two RRValidator and RFValidator objects.
"""

from __future__ import annotations

import json
import logging
import socket
import ssl
import typing as t
from dataclasses import asdict, dataclass, field
from pathlib import Path

import dns.edns
import dns.exception
import dns.flags
import dns.opcode
import dns.query
import dns.rcode
import dns.rdatatype
import dns.resolver

from dns_exporter.exceptions import ConfigError

if t.TYPE_CHECKING:  # pragma: no cover
    import urllib.parse
    from ipaddress import IPv4Address, IPv6Address


# get logger
logger = logging.getLogger(f"dns_exporter.{__name__}")

# the currently supported protocols in dns_exporter
valid_protocols = [
    "udp",
    "tcp",
    "udptcp",
    "dot",
    "doh",
    "doh3",
    "doq",
]


@dataclass
class RRValidator:
    """``dns_exporter.config.RRValidator`` defines the structure used in ``Config`` objects to validate response RRs.

    It is used in the ``validate_(answer|authority|additional)_rrs`` settings in the config.

    Like the parent ``dns_exporter.config.Config`` class it consists of a bunch of attributes and the
    ``dns_exporter.config.RRValidator.create()`` method which returns an instance of this class.

    Each Config object can have up to three (3) instances of this class, one for validating RRs in each section
    of the response:

        - ``answer``
        - ``authority``
        - ``additional``
    """

    fail_if_matches_regexp: list[str]
    """``fail_if_matches_regexp`` is a list of regular expressions, fail the query if one of them matches
    an RR."""

    fail_if_all_match_regexp: list[str]
    """``fail_if_all_match_regexp`` is a list of regular expressions, fail the query if all of them matches
    an RR."""

    fail_if_not_matches_regexp: list[str]
    """``fail_if_not_matches_regexp`` is a list of regular expressions, fail the query if one of them does not match
    an RR."""

    fail_if_none_matches_regexp: list[str]
    """``fail_if_none_matches_regexp`` is a list of regular expressions, fail the query if all of them does not match
    an RR."""

    fail_if_count_eq: int | None
    """``fail_if_count_eq`` is an integer, fail the query if the number of RRs matches this number."""

    fail_if_count_ne: int | None
    """``fail_if_count_ne`` is an integer, fail the query if the number of RRs is not equal to this number."""

    fail_if_count_lt: int | None
    """``fail_if_count_lt`` is an integer, fail the query if the number of RRs is smaller than this number."""

    fail_if_count_gt: int | None
    """``fail_if_count_gt`` is an integer, fail the query if the number of RRs is larger than this number."""

    @classmethod
    def create(  # noqa: PLR0913
        cls: type[RRValidator],
        fail_if_matches_regexp: list[str] | None = None,
        fail_if_all_match_regexp: list[str] | None = None,
        fail_if_not_matches_regexp: list[str] | None = None,
        fail_if_none_matches_regexp: list[str] | None = None,
        fail_if_count_eq: int | None = None,
        fail_if_count_ne: int | None = None,
        fail_if_count_lt: int | None = None,
        fail_if_count_gt: int | None = None,
    ) -> RRValidator:
        """Return an instance of the RRValidator class with values from the provided parameters.

        The values are used as-is.
        """
        # immutable defaults
        if fail_if_matches_regexp is None:
            fail_if_matches_regexp = []
        if fail_if_all_match_regexp is None:
            fail_if_all_match_regexp = []
        if fail_if_not_matches_regexp is None:
            fail_if_not_matches_regexp = []
        if fail_if_none_matches_regexp is None:
            fail_if_none_matches_regexp = []
        return cls(
            fail_if_matches_regexp=fail_if_matches_regexp,
            fail_if_all_match_regexp=fail_if_all_match_regexp,
            fail_if_not_matches_regexp=fail_if_not_matches_regexp,
            fail_if_none_matches_regexp=fail_if_none_matches_regexp,
            fail_if_count_eq=fail_if_count_eq,
            fail_if_count_ne=fail_if_count_ne,
            fail_if_count_lt=fail_if_count_lt,
            fail_if_count_gt=fail_if_count_gt,
        )


@dataclass
class RFValidator:
    """``dns_exporter.config.RFValidator`` defines the structure used in ``Config`` objects to validate response flags.

    It is used in the ``validate_response_flags`` setting in the config.

    Like the parent ``dns_exporter.config.Config`` class it consists of a bunch of attributes and the
    ``dns_exporter.config.RFValidator.create()`` method which returns an instance of this class.
    """

    fail_if_any_present: list[str]
    """``fail_if_any_present`` is a list of flags as strings, fail if any of them are present in the response."""

    fail_if_all_present: list[str]
    """``fail_if_all_present`` is a list of flags as strings, fail if all of them are present in the response."""

    fail_if_any_absent: list[str]
    """``fail_if_any_absent`` is a list of flags as strings, fail if any of them are absent from the response."""

    fail_if_all_absent: list[str]
    """``fail_if_all_present`` is a list of flags as strings, fail if all of them are absent from the response."""

    @classmethod
    def create(
        cls: type[RFValidator],
        fail_if_any_present: list[str] | None = None,
        fail_if_all_present: list[str] | None = None,
        fail_if_any_absent: list[str] | None = None,
        fail_if_all_absent: list[str] | None = None,
    ) -> RFValidator:
        """Return an instance of the RFValidator class with values from the provided parameters.

        The values are used as-is.
        """
        # immutable defaults
        if fail_if_any_present is None:
            fail_if_any_present = []
        if fail_if_all_present is None:
            fail_if_all_present = []
        if fail_if_any_absent is None:
            fail_if_any_absent = []
        if fail_if_all_absent is None:
            fail_if_all_absent = []
        return cls(
            fail_if_any_present=fail_if_any_present,
            fail_if_all_present=fail_if_all_present,
            fail_if_any_absent=fail_if_any_absent,
            fail_if_all_absent=fail_if_all_absent,
        )


@dataclass
class Config:
    """``dns_exporter.config.Config`` defines the primary config structure used in dns_exporter.

    The defaults for each config key are defined in the ``dns_exporter.config.Config.create()`` method.

    The ``dns_exporter.exporter.DNSExporter.modules`` dict consists of string keys and instances of this class as
    values.
    """

    name: str
    """str: The name of this config. It is mostly included in the class for convenience."""

    # required
    collect_ttl: bool
    """bool: Set this bool to ``True`` to enable collection of per-RR TTL metrics for the DNS query, ``False`` to not
    collect per-RR TTL metrics. Default is ``True``"""

    collect_ttl_rr_value_length: int
    """int: Limits the length of the ``rr_value`` label when collecing per-RR TTL metrics. Default is ``50``"""

    connection_reuse: bool
    """bool: Set this bool to ``True`` to keep and re-use sockets and sessions when possible. Default is ``False``"""

    edns: bool
    """bool: Set this bool to ``True`` to enable ``EDNS0`` for the DNS query, ``False`` to not use ``EDNS0``.
    Default is ``True``"""

    edns_do: bool
    """bool: Set this bool to ``True`` to set the ``EDNS0`` ``DO`` flag for the DNS query. Default is ``False``"""

    edns_nsid: bool
    """bool: Set this bool to ``True`` to set the ``EDNS0`` ``nsid`` option for the DNS query. Default is ``True``"""

    edns_bufsize: int
    """int: This int sets the ``EDNS0`` bufsize for the DNS query. Default is ``1232``"""

    edns_pad: int
    """int: This int sets the ``EDNS0`` padding size for the DNS query. Default is ``0``"""

    family: str
    """str: This string key must be set to either ``ipv6`` or ``ipv4``. It determines the address family used for the
    DNS query. Default is ``ipv6``"""

    protocol: str
    """str: This key must be set to one of ``udp``, ``tcp``, ``udptcp``, ``dot``, ``doh``, or ``doq``. It determines
    the protocol used for the DNS query. Default is ``udp``"""

    query_class: str
    """str: The query class used for this DNS query, typically ``IN`` but can also be ``CHAOS``. Default is ``IN``"""

    query_type: str
    """str: The query type used for this DNS query, like ``A`` or ``MX``. Default is ``A``"""

    proxy: urllib.parse.SplitResult | None
    """str: The proxy to use for this DNS query, for example ``socks5://127.0.0.1:5000``. Supported proxy types are
    SOCKS4, SOCKS5, and HTTP. Leave empty to use no proxy. Default is no proxy."""

    recursion_desired: bool
    """bool: Set this bool to ``True`` to set the ``RD`` flag in the DNS query. Default is ``True``"""

    timeout: float
    """float: This float determines how long the exporter will wait for a response before declaring the DNS query
    failed. Unit is seconds. Default is 5.0."""

    validate_answer_rrs: RRValidator
    """RRValidator: This object contains the validation config for the ``answer`` section of the response. Default
    is an empty ``RRValidator()``"""

    validate_authority_rrs: RRValidator
    """RRValidator: This object contains the validation config for the ``authority`` section of the response. Default
    is an empty ``RRValidator()``"""

    validate_additional_rrs: RRValidator
    """RRValidator: This object contains the validation config for the ``additional`` section of the response.
    Default is an empty ``RRValidator()``"""

    validate_response_flags: RFValidator
    """RFValidator: This object contains the validation config for the response flags. Default is an empty
    ``RFValidator()``"""

    valid_rcodes: list[str]
    """list[str]: A list of acceptable rcodes when validating the DNS response. Default is ``["NOERROR"]``."""

    verify_certificate: bool
    """bool: Set this bool to ``True`` to verify the certificate of the DNS server, set it to ``False`` to disable
    certificate verification. When enabled the default system CA can be overridden with the verify_certificate_path
    setting. Certificate validation is ignored for unencrypted protocols. Default is ``True``"""

    verify_certificate_path: str
    """bool: Set this to the path of a CA dir or CA file to override the default system CA when verifying certificates
    of encrypted DNS servers. Leave empty to use the default system CA path. Default is an empty string."""

    # optional settings (but required in final config)

    ip: IPv4Address | IPv6Address | None = field(
        default_factory=lambda: None,
    )
    """IPv4Address | IPv6Address | None: The IP to use instead of using IP or hostname from server. Default
    is ``None``"""

    server: urllib.parse.SplitResult | None = field(default_factory=lambda: None)
    """urllib.parse.SplitResult | None: The DNS server to use in parsed form. Default is ``None``"""

    query_name: str | None = field(default_factory=lambda: None)
    """str | None: The name to ask for in the DNS query. Default is ``None``"""

    def validate_bools(self) -> None:
        """Validate bools."""
        for key in [
            "collect_ttl",
            "connection_reuse",
            "edns",
            "edns_do",
            "edns_nsid",
            "recursion_desired",
            "verify_certificate",
        ]:
            # validate bools
            if not isinstance(getattr(self, key), bool):
                logger.error(f"Not a bool: {key}")
                raise ConfigError("invalid_request_config")

    def validate_integers(self) -> None:
        """Validate integers."""
        for key in ["collect_ttl_rr_value_length", "edns_bufsize", "edns_pad"]:
            if not getattr(self, key) >= 0:
                logger.error("Invalid integer")
                raise ConfigError("invalid_request_config")

    def validate_protocol(self) -> None:
        """Validate protocol."""
        if self.protocol not in valid_protocols:
            raise ConfigError(
                "invalid_request_protocol",
            )

    def validate_valid_qtypes(self) -> None:
        """Validate query_type (make sure it is supported in dnspython)."""
        valid_qtypes = [dns.rdatatype.to_text(t) for t in dns.rdatatype.RdataType]
        if self.query_type not in valid_qtypes:
            raise ConfigError(
                "invalid_request_query_type",
            )

    def validate_valid_rcodes(self) -> None:
        """Validate valid_rcodes."""
        all_rcodes = [dns.rcode.to_text(x) for x in dns.rcode.Rcode]
        invalid_rcodes = set(self.valid_rcodes).difference(all_rcodes)
        if invalid_rcodes:
            logger.error(f"Invalid rcodes used: {invalid_rcodes}")
            raise ConfigError(
                "invalid_request_config",
            )

    def validate_proxy(self) -> None:
        """Validate proxy."""
        if self.proxy and self.protocol in ["dot"]:
            # proxy support doesn't work for DoT for now
            logger.error(
                f"proxy not valid for protocol {self.protocol} see https://github.com/tykling/dns_exporter/issues/76"
            )
            raise ConfigError(
                "invalid_request_config",
            )

    def __post_init__(self) -> None:
        """Validate as much as possible."""
        # validate bools
        self.validate_bools()

        # validate integers
        self.validate_integers()

        # validate family
        if self.family not in ["ipv4", "ipv6"]:
            raise ConfigError("invalid_request_family")

        # validate protocol
        self.validate_protocol()

        # validate query_class
        if self.query_class not in ["IN", "CHAOS"]:
            raise ConfigError("invalid_request_query_class")

        # validate query_type
        self.validate_valid_qtypes()

        # validate valid_rcodes
        self.validate_valid_rcodes()

        # validate proxy
        self.validate_proxy()

        # validate ca path
        if self.verify_certificate_path:
            certpath = Path(self.verify_certificate_path)
            if not certpath.exists():
                raise ConfigError("invalid_request_path")

    @classmethod
    def create(  # noqa: PLR0913
        cls: type[Config],
        *,
        name: str,
        collect_ttl: bool = True,
        collect_ttl_rr_value_length: int = 50,
        connection_reuse: bool = False,
        edns: bool = True,
        edns_do: bool = False,
        edns_nsid: bool = True,
        edns_bufsize: int = 1232,
        edns_pad: int = 0,
        family: str = "ipv6",
        protocol: str = "udp",
        query_class: str = "IN",
        query_type: str = "A",
        recursion_desired: bool = True,
        proxy: urllib.parse.SplitResult | None = None,
        timeout: float = 5.0,
        validate_answer_rrs: RRValidator | None = None,
        validate_authority_rrs: RRValidator | None = None,
        validate_additional_rrs: RRValidator | None = None,
        validate_response_flags: RFValidator | None = None,
        valid_rcodes: list[str] | None = None,
        verify_certificate: bool = True,
        verify_certificate_path: str = "",
        ip: IPv4Address | IPv6Address | None = None,
        server: urllib.parse.SplitResult | None = None,
        query_name: str | None = None,
    ) -> Config:
        """Return an instance of the Config class with values from the provided parameters overriding the defaults."""
        logger.debug(f"creating config {name}...")
        # immutable defaults
        if valid_rcodes is None:
            valid_rcodes = ["NOERROR"]

        if validate_answer_rrs is None:
            validate_answer_rrs = RRValidator.create()

        if validate_authority_rrs is None:
            validate_authority_rrs = RRValidator.create()

        if validate_additional_rrs is None:
            validate_additional_rrs = RRValidator.create()

        if validate_response_flags is None:
            validate_response_flags = RFValidator.create()

        return cls(
            name=name,
            collect_ttl=collect_ttl,
            collect_ttl_rr_value_length=collect_ttl_rr_value_length,
            connection_reuse=connection_reuse,
            edns=edns,
            edns_do=edns_do,
            edns_nsid=edns_nsid,
            edns_bufsize=int(edns_bufsize),
            edns_pad=int(edns_pad),
            family=family,
            protocol=protocol,
            query_class=query_class.upper(),
            query_type=query_type.upper(),
            recursion_desired=recursion_desired,
            proxy=proxy,
            timeout=float(timeout),
            validate_answer_rrs=validate_answer_rrs,
            validate_authority_rrs=validate_authority_rrs,
            validate_additional_rrs=validate_additional_rrs,
            validate_response_flags=validate_response_flags,
            valid_rcodes=list(valid_rcodes),
            verify_certificate=verify_certificate,
            verify_certificate_path=verify_certificate_path,
            # fields with no defaults below here
            ip=ip,
            server=server,
            query_name=query_name,
        )

    def json(self) -> str:
        """Return a json version of the config. Mostly used in unit tests."""
        conf: dict[str, t.Any] = asdict(self)
        conf["ip"] = str(conf["ip"])
        conf["server"] = conf["server"].geturl()
        if conf["proxy"]:
            conf["proxy"] = conf["proxy"].geturl()
        return json.dumps(conf)

    def get_tls_context(self) -> ssl.SSLContext | bool:
        """Return a bool or ssl.SSLContext instance. Only used by protocol doh (which uses httpx)."""
        # is there a custom verify_certificate_path?
        if self.verify_certificate_path and self.verify_certificate:
            # verify with custom ca path, determine dir or file
            certpath = Path(self.verify_certificate_path)
            if certpath.is_dir():
                return ssl.create_default_context(capath=str(certpath), cafile=None, cadata=None)
            if certpath.is_file():
                return ssl.create_default_context(capath=None, cafile=str(certpath), cadata=None)
            # verify_certificate_path is neither dir or file, do not return a context
        # do cert verification?
        return self.verify_certificate

    def get_tls_verify(self) -> bool | str:
        """Return a bool or str for TLS verify args. Used by DoT, DoQ, DoH3."""
        if self.verify_certificate_path and self.verify_certificate:
            return self.verify_certificate_path
        return self.verify_certificate

    @property
    def verify(self) -> str | bool | ssl.SSLContext | None:
        """Get verify value suitable for different protocols."""
        if self.protocol == "doh":
            return self.get_tls_context()
        if self.protocol in ["dot", "doq", "doh3"]:
            return self.get_tls_verify()
        return None

    @property
    def dest(self) -> tuple[str, int]:
        """Return an IP, port tuple."""
        if t.TYPE_CHECKING:  # pragma: no cover
            assert self.server is not None
            assert isinstance(self.server.port, int)
        return (str(self.ip), self.server.port)

    @property
    def socket_family(self) -> socket.AddressFamily:
        """Return the socket family."""
        return socket.AF_INET if self.family == "ipv4" else socket.AF_INET6


class ConfigDict(t.TypedDict, total=False):
    """A TypedDict to help hold config dicts before they become Config objects.

    ``dns_exporter.config.ConfigDict`` behaves like a regular dict but works better with mypy
    because the individual keys has been annotated.

    ``dns_exporter.config.ConfigDict`` has all the same keys and types as the real final
    ``dns_exporter.config.Config`` object does.
    """

    collect_ttl: bool
    collect_ttl_rr_value_length: bool
    connection_reuse: bool
    edns: bool
    edns_do: bool
    edns_nsid: bool
    edns_bufsize: int
    edns_pad: int
    family: str
    protocol: str
    query_class: str
    query_type: str
    recursion_desired: bool
    timeout: float
    validate_answer_rrs: RRValidator
    validate_authority_rrs: RRValidator
    validate_additional_rrs: RRValidator
    validate_response_flags: RFValidator
    valid_rcodes: list[str]
    verify_certificate: bool
    verify_certificate_path: str
    ip: IPv4Address | IPv6Address | None
    server: urllib.parse.SplitResult | None
    query_name: str | None
    proxy: urllib.parse.SplitResult | None



"""Unit tests for proxy functionality."""

import logging
import sys

import pytest
import requests

###################################################################################


@pytest.mark.parametrize(
    ("protocol", "server"),
    [
        ("udp", "dns.google"),
        ("tcp", "dns.google"),
        pytest.param(
            "dot",
            "anycast.uncensoreddns.org",
            marks=pytest.mark.xfail(
                reason="Proxy not yet supported for DoT, see https://github.com/tykling/dns_exporter/issues/76"
            ),
        ),
        ("doh", "anycast.uncensoreddns.org"),
        ("doq", "dns-unfiltered.adguard.com"),
        ("doh3", "dns-unfiltered.adguard.com"),
    ],
)
@pytest.mark.xfail(
    sys.version_info[:2] == (3, 14), reason="Skip under Python 3.14 https://github.com/tykling/dns_exporter/issues/202"
)
def test_proxy(dns_exporter_example_config, proxy_server, protocol, server):
    """Test proxy functionality for all protocols."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": server,
            "family": "ipv4",
            "protocol": protocol,
            "proxy": "socks5://127.0.0.1",
        },
        timeout=5,
    )
    assert 'proxy="socks5://127.0.0.1:1080"' in r.text
    assert f'server="{protocol}://{server}:' in r.text


###################################################################################


@pytest.mark.xfail(
    sys.version_info[:2] == (3, 14), reason="Skip under Python 3.14 https://github.com/tykling/dns_exporter/issues/202"
)
@pytest.mark.parametrize("protocol", ["udp", "tcp", "dot", "doh", "doh3", "doq"])
def test_proxy_fail(dns_exporter_example_config, proxy_server, protocol):
    """Test proxy failure for all protocols."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "family": "ipv4",
            "protocol": protocol,
            "proxy": "socks5://127.0.0.1:1081",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


###################################################################################


def test_proxy_without_scheme(dns_exporter_example_config):
    """Trigger an invalid_request_proxy failure by providing a proxy without a scheme."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "proxy": "127.0.0.1:1080",
        },
        timeout=5,
    )
    assert r.status_code == 200, "non-200 returncode"


def test_proxy_unknown_scheme(dns_exporter_example_config):
    """Trigger an invalid_request_proxy failure by providing a proxy with an unknown scheme."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "proxy": "foo://127.0.0.1:1080",
        },
        timeout=5,
    )
    assert r.status_code == 200, "non-200 returncode"


def test_exporter_modules_none(caplog, exporter):
    """Make sure calling configure() with modules=None works."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    exporter.configure(modules=None)
    assert "0 module(s) loaded OK, total modules: 0." in caplog.text


@pytest.mark.xfail(
    sys.version_info[:2] == (3, 14), reason="Skip under Python 3.14 https://github.com/tykling/dns_exporter/issues/202"
)
def test_proxy_module(dns_exporter_example_config, proxy_server):
    """Test proxy functionality for udp protocol."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "family": "ipv4",
            "protocol": "udp",
            "module": "socks1080",
        },
        timeout=5,
    )
    assert 'proxy="socks5://127.0.0.1:1080"' in r.text
    assert 'server="udp://dns.google:53"' in r.text



"""``dns_exporter.version`` takes care of getting the package version from SCM or file.

The module contains no functions or methods and only a single module-level variable which is the version.
"""

import logging
from importlib.metadata import PackageNotFoundError, version

# get logger
logger = logging.getLogger(f"dns_exporter.{__name__}")

# get version number from package metadata if possible
try:
    __version__ = version("dns_exporter")
except PackageNotFoundError:  # pragma: no cover
    # package is not installed, get version from file
    try:
        from _version import version as __version__  # type: ignore[no-redef,import-not-found]
    except ImportError:
        # this must be a git checkout with no _version.py file, version unknown
        __version__: str = "0.0.0"  # type: ignore[no-redef]
logger.debug(f"Detected version running: {__version__}")



"""Unit tests for entrypoint.py."""

import time

import pytest

from dns_exporter.entrypoint import main
from dns_exporter.version import __version__

mockargs = [
    "-c",
    "dns_exporter/dns_exporter_example.yml",
    "-d",
    "-p",
    "25353",
]


def test_listen_port_busy(dns_exporter_example_config, caplog):
    """Test calling main() on a port which is already busy."""
    with pytest.raises(SystemExit):
        main(mockargs)
    time.sleep(2)
    assert "is in use?" in caplog.text


def test_version(capsys):
    """Make sure the -v command-line option returns the version."""
    with pytest.raises(SystemExit):
        main(["-v"])
    captured = capsys.readouterr()
    assert __version__ in captured.out


def test_socket_cache_max_age_option(dns_exporter_example_config, caplog):
    """Make sure the --connection-max-age-seconds command-line option works."""
    with pytest.raises(SystemExit):
        main([*mockargs, "--connection-max-age-seconds", "1234"])
    assert "SocketCache initialised with max. age 1234 seconds" in caplog.text


def test_socket_cache_max_idle_option(dns_exporter_example_config, caplog):
    """Make sure the --connection-max-idle-seconds command-line option works."""
    with pytest.raises(SystemExit):
        main([*mockargs, "--connection-max-idle-seconds", "1234"])
    assert "and max. idle 1234 seconds" in caplog.text


def test_socket_cache_cleanup_interval_option(dns_exporter_example_config, caplog):
    """Make sure the --connection-cleanup-interval-seconds command-line option works."""
    with pytest.raises(SystemExit):
        main([*mockargs, "--connection-cleanup-interval-seconds", "1234"])
    assert "and housekeeping interval 1234 seconds" in caplog.text



"""dns_exporter tests for the config module."""

from ipaddress import IPv4Address

import pytest
import requests

from dns_exporter.config import Config, ConfigDict, ConfigError


def test_nonbool_bool(exporter):
    """Test a bool which is not a bool."""
    with pytest.raises(ConfigError):
        exporter.prepare_config(ConfigDict(edns_do=42))
    prepared = exporter.prepare_config(ConfigDict(edns_do=True))
    prepared["edns_do"] = 42
    with pytest.raises(ConfigError):
        Config.create(name="test", **prepared)


def test_negative_int(exporter):
    """Test a negative int."""
    prepared = exporter.prepare_config(ConfigDict(edns_pad=-1))
    with pytest.raises(ConfigError):
        Config.create(name="test", **prepared)


def test_invalid_protocol(exporter):
    """Test with an unknown protocol."""
    prepared = exporter.prepare_config(ConfigDict(protocol="sctp"))
    with pytest.raises(ConfigError):
        Config.create(name="test", **prepared)


def test_invalid_class(exporter):
    """Test with an unknown class."""
    prepared = exporter.prepare_config(ConfigDict(query_class="OUT"))
    with pytest.raises(ConfigError):
        Config.create(name="test", **prepared)


def test_invalid_qtype(exporter):
    """Test with an unknown qtype."""
    prepared = exporter.prepare_config(ConfigDict(query_type="B"))
    with pytest.raises(ConfigError):
        Config.create(name="test", **prepared)


def test_invalid_timeout(exporter):
    """Test with an invalid float."""
    with pytest.raises(ConfigError):
        exporter.prepare_config(ConfigDict(timeout="timein"))


def test_invalid_rrvalidator(exporter):
    """Test with something not an RRValidator object."""
    with pytest.raises(ConfigError):
        exporter.prepare_config(ConfigDict(validate_answer_rrs=42))


def test_invalid_rfvalidator(exporter):
    """Test with something not an RFValidator object."""
    with pytest.raises(ConfigError):
        exporter.prepare_config(ConfigDict(validate_response_flags=42))


def test_invalid_rcode(exporter):
    """Test with an invalid RCODE."""
    with pytest.raises(ConfigError):
        prepared = exporter.prepare_config(ConfigDict(valid_rcodes=42))
    prepared = exporter.prepare_config(ConfigDict(valid_rcodes="YESERROR"))
    with pytest.raises(ConfigError):
        Config.create(name="test", **prepared)


def test_edns_string(exporter):
    """Test when edns is a string."""
    prepared = exporter.prepare_config(ConfigDict(edns="true"))
    c = Config.create(name="test", **prepared)
    assert c.edns is True


def test_edns_do_false(exporter):
    """Test when edns_do is set to a string."""
    prepared = exporter.prepare_config(ConfigDict(edns_do="false"))
    c = Config.create(name="test", **prepared)
    assert c.edns_do is False


def test_edns_do_true(exporter):
    """Test when edns_do is set to a string."""
    prepared = exporter.prepare_config(ConfigDict(edns_do="true"))
    c = Config.create(name="test", **prepared)
    assert c.edns_do is True


def test_rd_true(exporter):
    """Test when recursion_desired is set to a string."""
    prepared = exporter.prepare_config(ConfigDict(recursion_desired="true"))
    c = Config.create(name="test", **prepared)
    assert c.recursion_desired is True


def test_rd_false(exporter):
    """Test when recursion_desired is set to a string."""
    prepared = exporter.prepare_config(ConfigDict(recursion_desired="false"))
    c = Config.create(name="test", **prepared)
    assert c.recursion_desired is False


def test_proxy_for_unsupported_protocol(exporter):
    """Test proxy with a protocol not supported."""
    prepared = exporter.prepare_config(
        ConfigDict(protocol="dot", proxy="socks5://127.0.0.1"),
    )
    with pytest.raises(ConfigError):
        Config.create(name="test", **prepared)


def test_prepare_config_ip_real_ipaddress(caplog, exporter):
    """Make sure a ConfigDict with a real ipaddress.IPv4Address or ipaddress.IPv6Address object works as intended."""
    indict = ConfigDict(ip=IPv4Address("127.0.0.1"))
    outdict = exporter.prepare_config_ip(config=indict)
    assert isinstance(outdict["ip"], IPv4Address)


def test_prepare_config_ip_badtype(caplog, exporter):
    """Make sure a ConfigDict with an integer ip fails as intended."""
    indict = ConfigDict(ip=42)
    with pytest.raises(TypeError):
        exporter.prepare_config_ip(config=indict)


def test_edns_bufsize_nonint(exporter, caplog):
    """Test when edns_bufsize is set to a wrong type."""
    with pytest.raises(ConfigError):
        exporter.prepare_config(ConfigDict(edns_bufsize=[1, 2, 3]))
    assert "Unable to validate integer for key edns_bufsize" in caplog.text


def test_timeout_nofloat(exporter, caplog):
    """Test when timeout is set to neither a string or a float."""
    with pytest.raises(ConfigError):
        exporter.prepare_config(ConfigDict(timeout=["foo"]))
    assert "Unable to validate float for key timeout" in caplog.text


def test_int_server(exporter, caplog):
    """Test when server is set to a wrong type."""
    with pytest.raises(ConfigError):
        exporter.prepare_config(ConfigDict(server=42, protocol="udp"))


def test_int_proxy(exporter, caplog):
    """Test when proxy is set to a wrong type."""
    with pytest.raises(ConfigError):
        exporter.prepare_config(ConfigDict(proxy=42))


# test querystring configuration


def test_config_querystring_rcode(dns_exporter_example_config):
    """Test setting rcode in request querystring."""
    r = requests.get(
        "http://127.0.0.1:25353/config",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "valid_rcodes": "NXDOMAIN",
        },
        timeout=5,
    )
    config = r.json()
    assert config["valid_rcodes"] == ["NXDOMAIN"]


def test_config_querystring_rcode_2(dns_exporter_example_config):
    """Test setting multiple rcodes in request querystring."""
    r = requests.get(
        "http://127.0.0.1:25353/config",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "valid_rcodes": "NXDOMAIN,NOERROR",
        },
        timeout=5,
    )
    config = r.json()
    assert config["valid_rcodes"] == ["NXDOMAIN", "NOERROR"]



"""``dns_exporter.exporter`` contains the DNSExporter class.

The config.py module contains configuration related stuff, metrics.py
contains the metric definitions, collector.py has the Collector and
and this exporter.py module contains most of the rest of the code.

    - Repository: https://github.com/tykling/dns_exporter
    - Pypi: https://pypi.org/project/dns-exporter/
    - Docs: https://dns-exporter.readthedocs.io/en/latest/

Made with love by Thomas Steen Rasmussen/Tykling, 2023.
"""
# mypy: disable-error-code="literal-required"

from __future__ import annotations

import ipaddress
import logging
import random
import socket
import urllib.parse
from dataclasses import asdict
from ipaddress import IPv4Address, IPv6Address
from typing import TYPE_CHECKING, Literal

import dns.edns
import dns.exception
import dns.flags
import dns.opcode
import dns.query
import dns.rdatatype
import dns.resolver
import socks  # type: ignore[import-untyped]
from prometheus_client import CollectorRegistry, MetricsHandler, exposition

from dns_exporter.collector import DNSCollector, FailCollector
from dns_exporter.config import Config, ConfigDict, RFValidator, RRValidator
from dns_exporter.exceptions import ConfigError
from dns_exporter.metrics import QTIME_LABELS, dnsexp_http_requests_total, dnsexp_http_responses_total
from dns_exporter.socket_cache import SocketCache
from dns_exporter.version import __version__

if TYPE_CHECKING:  # pragma: no cover
    from dns.message import QueryMessage
    from prometheus_client.registry import RestrictedRegistry

logger = logging.getLogger(f"dns_exporter.{__name__}")
socket_cache = SocketCache()

INDEX = """<!DOCTYPE html>
<html lang="en">
<head><title>DNS Exporter</title></head>
<body>
<h1>DNS Exporter</h1>
<p>Visit <a href="/query?server=dns.google&protocol=doh&query_name=example.com">/query?server=dns.google&protocol=doh&query_name=example.com</a> to do a DNS query and see metrics.</p>
<p>To debug configuration issues replace /query with /config to see the effective parsed configuration:</p>
<p>Visit <a href="/config?server=dns.google&protocol=doh&query_name=example.com">/config?server=dns.google&protocol=doh&query_name=example.com</a> to see the final scrape config without doing a DNS query.</p>
<p>Visit <a href="/metrics">/metrics</a> to see metrics for the dns_exporter itself.</p>
</body>
</html>"""  # noqa: E501


class DNSExporter(MetricsHandler):
    """Primary dns_exporter class.

    MetricsHandler subclass for incoming scrape requests. Initiated on each
    request as a handler by http.server.HTTPServer().

    The configure() classmethod can optionally be called to load modules before use.

    Attributes:
    -----------
        modules: A dict of dns_exporter.config.Config instances to be used in scrape requests.

    """

    __version__ = __version__

    # the modules key is populated by configure() before the class is initialised
    modules: dict[str, Config] | None = None

    # bool to control the inclusion of the connection label
    connection_label: bool = False

    @classmethod
    def prepare_config_rrvalidators(
        cls,
        config: ConfigDict,
    ) -> ConfigDict:
        """Parse and create RRValidator objects for the config."""
        tmp: ConfigDict = {}

        answer: Literal["validate_answer_rrs"] = "validate_answer_rrs"
        authority: Literal["validate_authority_rrs"] = "validate_authority_rrs"
        additional: Literal["validate_additional_rrs"] = "validate_additional_rrs"
        try:
            # create RRValidator objects
            for validator in [answer, authority, additional]:
                if validator in config:
                    if isinstance(config[validator], dict):
                        tmp[validator] = RRValidator.create(**config[validator])
                    elif isinstance(config[validator], RRValidator):
                        tmp[validator] = config[validator]
                    else:
                        # unsupported type
                        raise TypeError(validator)  # noqa: TRY301
        except TypeError as e:
            logger.exception("Unable to create RRValidator object")
            raise ConfigError("invalid_request_config") from e
        return tmp

    @classmethod
    def prepare_config_rfvalidator(
        cls,
        config: ConfigDict,
    ) -> ConfigDict:
        """Parse and create RFValidator object for the config."""
        tmp: ConfigDict = {}
        try:
            # create RFValidator
            if "validate_response_flags" in config:
                if isinstance(config["validate_response_flags"], dict):
                    tmp["validate_response_flags"] = RFValidator.create(**config["validate_response_flags"])
                elif isinstance(config["validate_response_flags"], RFValidator):
                    tmp["validate_response_flags"] = config["validate_response_flags"]
                else:
                    # unsupported type
                    raise TypeError("validate_response_flags")  # noqa: TRY301
        except TypeError as e:
            logger.exception("Unable to create validator object")
            raise ConfigError("invalid_request_config") from e
        return tmp

    @classmethod
    def prepare_config_ip(
        cls,
        config: ConfigDict,
    ) -> ConfigDict:
        """Parse IP address if needed."""
        tmp: ConfigDict = {}
        key = "ip"
        if config.get(key):
            if isinstance(config["ip"], IPv4Address | IPv6Address):
                # use as-is
                tmp[key] = config[key]
            elif isinstance(config[key], str):
                # make an ip object
                try:
                    tmp[key] = ipaddress.ip_address(config[key])
                except ValueError as e:
                    logger.exception(f"Unable to parse IP address {config[key]}")
                    raise ConfigError("invalid_request_ip") from e
            else:
                # unsupported type
                raise TypeError(key)
        return tmp

    @classmethod
    def prepare_config_integers(
        cls,
        config: ConfigDict,
    ) -> ConfigDict:
        """Parse and create integer objects for the config."""
        tmp: ConfigDict = {}
        collect_ttl_rr_value_length: Literal["collect_ttl_rr_value_length"] = "collect_ttl_rr_value_length"
        edns_bufsize: Literal["edns_bufsize"] = "edns_bufsize"
        edns_pad: Literal["edns_pad"] = "edns_pad"
        try:
            for key in [edns_bufsize, edns_pad, collect_ttl_rr_value_length]:
                if key in config:
                    if isinstance(config[key], str):
                        tmp[key] = int(config[key])
                    elif isinstance(config[key], int):
                        # use as-is
                        tmp[key] = config[key]
                    else:
                        # unsupported type
                        raise TypeError(key)  # noqa: TRY301
        except (ValueError, TypeError) as e:
            logger.exception(f"Unable to validate integer for key {key}")
            raise ConfigError("invalid_request_config") from e
        return tmp

    @classmethod
    def prepare_config_floats(
        cls,
        config: ConfigDict,
    ) -> ConfigDict:
        """Parse and create float objects for the config."""
        tmp: ConfigDict = {}
        # use literals for TypedDict keys to make mypy happy
        timeout: Literal["timeout"] = "timeout"
        try:
            for key in [timeout]:
                if key in config:
                    if isinstance(config[key], str | int):
                        tmp[key] = float(config[key])
                    elif isinstance(config[key], float):
                        # use as-is
                        tmp[key] = config[key]
                    else:
                        # unsupported type
                        raise TypeError(key)  # noqa: TRY301
        except (ValueError, TypeError) as e:
            logger.exception(f"Unable to validate float for key {key}")
            raise ConfigError("invalid_request_config") from e
        return tmp

    @classmethod
    def prepare_config_bools(
        cls,
        config: ConfigDict,
    ) -> ConfigDict:
        """Parse and create bool objects for the config."""
        tmp: ConfigDict = {}
        # use literals for TypedDict keys to make mypy happy
        collect_ttl: Literal["collect_ttl"] = "collect_ttl"
        connection_reuse: Literal["connection_reuse"] = "connection_reuse"
        edns: Literal["edns"] = "edns"
        edns_do: Literal["edns_do"] = "edns_do"
        recursion_desired: Literal["recursion_desired"] = "recursion_desired"
        verify_certificate: Literal["verify_certificate"] = "verify_certificate"
        try:
            for key in [collect_ttl, connection_reuse, edns, edns_do, recursion_desired, verify_certificate]:
                if key not in config:
                    continue
                if isinstance(config[key], str):
                    # evaluate string as true if it feels truthy
                    tmp[key] = config[key].lower() in ("true", "t", "yes", "y")
                elif isinstance(config[key], bool):
                    # use as-is
                    tmp[key] = config[key]
                else:
                    # unsupported type
                    raise TypeError(key)  # noqa: TRY301
        except (ValueError, TypeError) as e:
            logger.exception(f"Unable to validate bool for key {key}")
            raise ConfigError("invalid_request_config") from e
        return tmp

    @classmethod
    def prepare_config_server(
        cls,
        config: ConfigDict,
    ) -> ConfigDict:
        """Parse server into a SplitResult and return it."""
        tmp: ConfigDict = {}
        key: Literal["server"] = "server"
        try:
            if config.get(key):
                server = config[key]
                if isinstance(server, str):
                    # parse server into a SplitResult
                    tmp[key] = cls.parse_server(
                        server=server,
                        protocol=config.get("protocol", "udp"),
                    )
                elif isinstance(server, urllib.parse.SplitResult):
                    # use as-is
                    tmp[key] = server
                else:
                    # unsupported type
                    raise TypeError(key)  # noqa: TRY301
        except TypeError as e:
            logger.exception("Unable to parse server")
            raise ConfigError("invalid_request_server") from e
        return tmp

    @classmethod
    def prepare_config_proxy(
        cls,
        config: ConfigDict,
    ) -> ConfigDict:
        """Parse proxy into a SplitResult and return it."""
        tmp: ConfigDict = {}
        key: Literal["proxy"] = "proxy"
        try:
            if config.get(key):
                proxy = config[key]
                if isinstance(proxy, str):
                    if "://" not in proxy:
                        logger.error("No scheme in proxy")
                        raise ConfigError("invalid_request_proxy")

                    # parse proxy into a SplitResult
                    splitresult = urllib.parse.urlsplit(proxy)
                    if not splitresult.scheme or splitresult.scheme.upper() not in socks.PROXY_TYPES:
                        logger.error(f"Invalid proxy scheme {splitresult}")
                        raise ConfigError("invalid_request_proxy")

                    # make port explicit
                    if splitresult.port is None:
                        # SOCKS4 and SOCKS5 default to port 1080
                        port = 8080 if splitresult.scheme == "http" else 1080
                        splitresult = splitresult._replace(
                            netloc=f"{splitresult.netloc}:{port}",
                        )

                    # keep only scheme and netloc
                    tmp[key] = urllib.parse.urlsplit(
                        splitresult.scheme + "://" + splitresult.netloc,
                    )
                    logger.debug(f"Using proxy {splitresult.geturl()!s}")
                elif isinstance(proxy, urllib.parse.SplitResult):
                    # use as-is
                    tmp[key] = proxy
                else:
                    # unsupported type
                    raise TypeError(key)  # noqa: TRY301
        except TypeError as e:
            logger.exception("Unable to parse proxy")
            raise ConfigError("invalid_request_proxy") from e
        return tmp

    @classmethod
    def prepare_config_lists(
        cls,
        config: ConfigDict,
    ) -> ConfigDict:
        """Parse and create list objects for the config."""
        tmp: ConfigDict = {}
        # use literals for TypedDict keys to make mypy happy
        valid_rcodes: Literal["valid_rcodes"] = "valid_rcodes"
        try:
            for key in [valid_rcodes]:
                if key in config:
                    if isinstance(config[key], str):
                        tmp[key] = str(config[key]).split(",")
                    elif isinstance(config[key], list):
                        # use as-is
                        tmp[key] = config[key]
                    else:
                        # unsupported type
                        raise TypeError(key)  # noqa: TRY301
        except (ValueError, TypeError) as e:
            logger.exception(f"Unable to validate list for key {key}")
            raise ConfigError("invalid_request_config") from e
        return tmp

    @classmethod
    def prepare_config(cls, config: ConfigDict) -> ConfigDict:
        """Make sure the configdict has the right types and objects.

        This method is called from:
          - DNSExporter.configure() (before class initialisation, optional)
          - During each scrape request

        Args:
        -----
            config: A ConfigDict instance

        Returns:
        --------
            A ConfigDict instance

        Raises:
        -------
            ConfigError: If any issues are found with the ConfigDict

        """
        # parse and create RRValidator objects
        config.update(cls.prepare_config_rrvalidators(config))
        # parse and create RFValidator object
        config.update(cls.prepare_config_rfvalidator(config))
        # parse and create IP object
        config.update(cls.prepare_config_ip(config))
        # validate integer keys
        config.update(cls.prepare_config_integers(config))
        # validate float keys
        config.update(cls.prepare_config_floats(config))
        # validate bool keys
        config.update(cls.prepare_config_bools(config))
        # validate list keys
        config.update(cls.prepare_config_lists(config))
        # parse server
        config.update(cls.prepare_config_server(config))
        # parse proxy
        config.update(cls.prepare_config_proxy(config))
        # all done
        return config

    def validate_config(self) -> None:
        """Validate various aspects of the config."""
        # make sure the server is valid, resolve ip if needed
        self.validate_server_ip()

        # make sure there is a query_name in the config
        if not self.config.query_name:
            raise ConfigError("invalid_request_query_name")

    @staticmethod
    def handle_failure(fail_registry: CollectorRegistry, failure: str, labels: dict[str, str]) -> None:
        """Handle various failure cases that can occur before the DNSCollector is called."""
        logger.debug(f"Initialising FailCollector to handle failure: {failure}")
        fail_collector = FailCollector(failure_reason=failure, labels=labels)
        logger.debug("Registering FailCollector in dnsexp_registry")
        fail_registry.register(fail_collector)

    def build_final_config(self, qs: dict[str, str]) -> None:
        """Construct the final effective scrape config from defaults and values from the querystring."""
        # first get the defaults
        config = ConfigDict(**asdict(Config.create(name="defaults")))  # type: ignore[typeddict-item]

        # if a module is specified in the querystring apply it first
        if "module" in qs:
            if self.modules is None or qs["module"] not in self.modules:
                raise ConfigError("invalid_request_module")
            config.update(asdict(self.modules[qs["module"]]))
            del qs["module"]

        # and the querystring from the scrape request has highest precedence
        config.update(qs)

        # prepare config dict
        config = self.prepare_config(config)
        # the final config has the name "final"
        del config["name"]  # type: ignore[typeddict-item]

        # create the config object
        try:
            self.config = Config.create(name="final", **config)
        except TypeError as e:
            logger.exception(
                "Exception while creating config - invalid field specified?",
            )
            raise ConfigError("invalid_request_config") from e

        # validate config
        self.validate_config()

        logger.debug(f"Final scrape configuration: {self.config}")

    @classmethod
    def configure(
        cls,
        modules: dict[str, ConfigDict] | None = None,
    ) -> bool:
        """Validate and create Config objects.

        Takes a dict of ConfigDict objects and runs cls.prepare_config() on each
        before creating a Config object and adding it to cls.modules

        If an error is encountered the process stops, but modules loaded until the
        failure can still be used in cls.modules.

        Args:
        -----
            modules: A dict of names and corresponding ConfigDict objects.

        Returns:
        --------
            bool: True if all ConfigDict objects was validated and loaded OK, False
                  if an error was encountered.
        """
        prepared: ConfigDict | None
        if modules is None:
            modules = {}
        if cls.modules is None:
            cls.modules = {}
        for name, config in modules.items():
            try:
                prepared = cls.prepare_config(ConfigDict(**config))
            except ConfigError:
                logger.exception(f"There was an issue while preparing config {name}")
                return False
            try:
                cls.modules[name] = Config.create(name=name, **prepared)
            except TypeError:
                logger.exception(f"Unable to parse config {prepared}")
                return False
            except ConfigError:
                logger.exception(
                    f"Invalid value found while building config {prepared}",
                )
                return False

        logger.info(
            f"{len(modules)} module(s) loaded OK, total modules: {len(cls.modules)}.",
        )
        return True

    @staticmethod
    def parse_server(server: str, protocol: str) -> urllib.parse.SplitResult:
        """Parse the server, add scheme (to make urllib.parse play ball), make port explicit.

        The server at this point can be:
          - a v4 IP
          - a v6 IP
          - a v4 ip:port
          - a v6 ip:port
          - a hostname
          - a hostname:port
          - a https:// url with an IP and no port
          - a https:// url with an IP:port
          - a https:// url with a hostname and no port
          - a https:// url with a hostname:port

        In the DoH/DoH3 https:// cases the url can be with or without a path.

        Parse it with urllib.parse.urlsplit, add explicit port if needed, and return the result.
        """
        # make sure we always have a scheme to help the parser
        if "://" not in server:
            server = f"{protocol}://{server}"
        # parse the string
        splitresult = urllib.parse.urlsplit(server)
        # make sure scheme is the dns_exporter internal protocol identifier (not https://)
        splitresult = splitresult._replace(scheme=protocol)
        if protocol in ["doh", "doh3"] and not splitresult.path:
            # use the default DoH path
            splitresult = splitresult._replace(path="/dns-query")
        # is there an explicit port in the configured server url? use default if not.
        if splitresult.port is None:
            if protocol in ["udp", "tcp", "udptcp"]:
                # plain DNS
                port = 53
            elif protocol in ["dot", "doq"]:
                # DoT and DoQ
                port = 853
            else:
                # DoH or DoH3
                port = 443
            logger.debug(
                f"No explicit port in configured server, using default for protocol {protocol}: {port}",
            )
            splitresult = splitresult._replace(netloc=f"{splitresult.netloc}:{port}")
        # return the parsed server
        logger.debug(f"Using server {splitresult.geturl()!s}")
        return splitresult

    def validate_server_ip(self) -> None:
        """Validate the server and resolve IP if needed."""
        # is there a server?
        if not self.config.server:
            logger.error("No server found in config")
            raise ConfigError("invalid_request_server")

        # is there already an IP in the config?
        if self.config.ip:
            logger.debug(f"checking ip {self.config.ip} of type {type(self.config.ip)}")

            # make sure the ip matches the configured address family
            if not self.check_ip_family(ip=self.config.ip, family=self.config.family):
                raise ConfigError("invalid_request_ip")

            # self.config.server.hostname can be either a hostname or an ip,
            # if it is an ip make sure there is no conflict with ip arg
            try:
                serverip = ipaddress.ip_address(str(self.config.server.hostname))
                if serverip != self.config.ip:
                    raise ConfigError("invalid_request_ip")
            except ValueError:
                # server host is a hostname not an ip,
                # the hostname will NOT be resolved, since there already is an ip to use
                pass
            method = "from config"
        else:
            try:
                # server host might be an ip, attempt to parse it as such
                self.config.ip = ipaddress.ip_address(str(self.config.server.hostname))
            except ValueError:
                # there is no ip in the config, need to get ip by resolving server in dns
                resolved = self.resolve_ip_getaddrinfo(
                    hostname=str(self.config.server.hostname),
                    family=str(self.config.family),
                )
                self.config.ip = ipaddress.ip_address(resolved)
            method = f"resolved from {self.config.server.hostname}"

        logger.debug(
            f"Using server IP {self.config.ip} ({method}) for the DNS server connection",
        )

    @staticmethod
    def check_ip_family(ip: IPv4Address | IPv6Address, family: str) -> bool:
        """Make sure the IP matches the address family."""
        return (ip.version == 4 and family == "ipv4") or (ip.version == 6 and family == "ipv6")  # noqa: PLR2004

    def resolve_ip_getaddrinfo(self, hostname: str, family: str) -> str:
        """Resolve the IP of a DNS server hostname."""
        logger.debug(
            f"resolve_ip_getaddrinfo() called with hostname {hostname} and family {family}",
        )
        try:
            if family == "ipv4":
                logger.debug(f"doing getaddrinfo for hostname {hostname} for ipv4")
                result = socket.getaddrinfo(hostname, 0, family=socket.AF_INET)
                return str(random.choice(result)[4][0])  # noqa: S311
            logger.debug(f"doing getaddrinfo for hostname {hostname} for ipv6")
            result = socket.getaddrinfo(hostname, 0, family=socket.AF_INET6)
            return str(random.choice(result)[4][0])  # noqa: S311
        except socket.gaierror as e:
            logger.exception("Unable to resolve server")
            raise ConfigError("invalid_request_server") from e

    def parse_querystring(self) -> tuple[urllib.parse.SplitResult, dict[str, str]]:
        """Parse the incoming url and then the querystring."""
        # parse incoming request
        url = urllib.parse.urlsplit(self.path)
        parsed_qs = urllib.parse.parse_qs(url.query)
        # querystring values are all lists when returned from parse_qs(),
        # so take the first item only (since multiple values are not supported)
        # behold, a valid usecase for dict comprehension!
        qs: dict[str, str] = {k: v[0] for k, v in parsed_qs.items()}
        return url, qs

    def handle_query_request(self) -> None:
        """Handle incoming HTTP GET requests to /query or /config."""
        logger.debug(
            f"Got {self.url.path} request from client {self.client_address}",
        )
        logger.debug(
            "Initialising CollectorRegistry dnsexp_registry and fail_registry",
        )
        dnsexp_registry = CollectorRegistry()
        self.fail_registry = CollectorRegistry()

        # begin labels dict
        self.labels: dict[str, str] = {}
        for key in QTIME_LABELS:
            # default all labels to the string "none"
            self.labels[key] = "none"

        # build and validate configuration for this scrape from defaults, config file and request querystring
        try:
            self.build_final_config(qs=self.qs)
        except ConfigError as E:
            self.handle_failure(self.fail_registry, str(E), labels=self.labels)
            # something is wrong with the config, send error response and bail out
            self.send_metric_response(registry=self.fail_registry, query=self.qs)
            return

        # if this is a config check return now
        if self.url.path == "/config":
            logger.debug("returning config")
            self.send_response(200)
            self.end_headers()
            self.wfile.write(self.config.json().encode("utf-8"))
            return

        # config is ready for action, begin the labels dict
        self.labels.update(
            {
                "server": str(self.config.server.geturl()),  # type: ignore[union-attr]
                "ip": str(self.config.ip),
                "port": str(self.config.server.port),  # type: ignore[union-attr]
                "protocol": str(self.config.protocol),
                "family": str(self.config.family),
                "proxy": str(self.config.proxy.geturl()) if self.config.proxy else "none",
                "query_name": str(self.config.query_name),
                "query_type": str(self.config.query_type),
            }
        )

        # get query message
        q = get_query(config=self.config)

        # register the DNSCollector in dnsexp_registry
        dns_collector = DNSCollector(config=self.config, query=q, labels=self.labels)
        dnsexp_registry.register(dns_collector)
        # send the response (which triggers the collect)
        logger.debug("Returning DNS query metrics")
        self.send_metric_response(registry=dnsexp_registry, query=self.qs)

    def do_GET(self) -> None:  # noqa: N802
        """Handle incoming HTTP GET requests."""
        # parse the scrape request url and querystring
        self.url, self.qs = self.parse_querystring()
        logger.debug(
            f"Got HTTP request for {self.url.geturl()} - parsed qs is {self.qs}",
        )
        # increase the persistent http request metric
        dnsexp_http_requests_total.labels(path=self.url.path).inc()

        # /query is for doing a DNS query, it returns metrics about just that one dns query
        if self.url.path in ["/query", "/config"]:
            self.handle_query_request()

        # this endpoint exposes metrics about the exporter itself and the python process
        elif self.url.path == "/metrics":
            logger.debug("Returning exporter metrics for request to /metrics")
            socket_cache.update_metrics()
            self.send_metric_response(registry=self.registry, query=self.qs)  # type: ignore[arg-type]

        # the root just returns a bit of informational html
        elif self.url.path == "/":
            # return a basic index page
            logger.debug("Returning index page for request to /")
            self.send_response(200)
            self.send_header("Content-Length", str(len(INDEX.encode("utf-8"))))
            self.end_headers()
            self.wfile.write(INDEX.encode("utf-8"))
            dnsexp_http_responses_total.labels(path="/", response_code=200).inc()

        # unknown endpoint
        else:
            logger.debug(f"Unknown endpoint '{self.url.path}' returning 404")
            self.send_response(404)
            msg = b"404 not found"
            self.send_header("Content-Length", str(len(msg)))
            self.end_headers()
            self.wfile.write(msg)
            dnsexp_http_responses_total.labels(
                path=self.url.path,
                response_code=404,
            ).inc()

    def send_metric_response(
        self,
        registry: CollectorRegistry | RestrictedRegistry,
        query: dict[str, str],
    ) -> None:
        """Bake and send output from the provided registry and querystring."""
        # Bake output
        status, headers, output = exposition._bake_output(  # type: ignore[no-untyped-call]  # noqa: SLF001
            registry=registry,
            accept_header=self.headers.get("Accept"),
            accept_encoding_header=self.headers.get("Accept-Encoding"),
            params=query,
            disable_compression=False,
        )
        headers.append(("Content-Length", str(len(output))))
        # Return output
        self.send_response(int(status.split(" ")[0]))
        for header in headers:
            self.send_header(*header)
        self.end_headers()
        self.wfile.write(output)
        dnsexp_http_responses_total.labels(path=self.url.path, response_code=200).inc()


def get_query(config: Config) -> QueryMessage:
    """Build and return the dns.message.QueryMessage object."""
    # prepare query
    qname = dns.name.from_text(str(config.query_name))
    q = dns.message.make_query(
        qname=qname,
        rdtype=str(config.query_type),
        rdclass=config.query_class,
    )

    # use EDNS?
    if config.edns:
        # use edns
        ednsargs: dict[
            str,
            str | int | bool | list[dns.edns.GenericOption],
        ] = {"options": []}
        # use the DO bit?
        if config.edns_do:
            ednsargs["ednsflags"] = dns.flags.DO
        # use nsid?
        if config.edns_nsid:
            ednsargs["options"].append(  # type: ignore[union-attr]
                dns.edns.GenericOption(dns.edns.NSID, ""),
            )
        # set bufsize/payload?
        if config.edns_bufsize:
            # dnspython calls bufsize "payload"
            ednsargs["payload"] = int(config.edns_bufsize)
        # set edns padding?
        if config.edns_pad:
            ednsargs["options"].append(  # type: ignore[union-attr]
                dns.edns.GenericOption(
                    dns.edns.PADDING,
                    bytes(int(config.edns_pad)),
                ),
            )
        # enable edns with the chosen options
        q.use_edns(edns=0, **ednsargs)  # type: ignore[arg-type]
        logger.debug(f"using edns options {ednsargs}")
    else:
        # do not use edns
        q.use_edns(edns=False)
        logger.debug("not using edns")

    # set RD flag?
    if config.recursion_desired:
        q.flags |= dns.flags.RD
    else:
        q.flags &= ~dns.flags.RD

    # go
    return q



"""Unit tests for ``dns_exporter.socket_cache`` related functionality."""

import logging
import socket
import ssl
import subprocess
import sys
import time
from concurrent.futures import ThreadPoolExecutor

import httpx
import pytest
import requests
from dns.quic._sync import SyncQuicConnection

from dns_exporter.config import Config, ConfigDict
from dns_exporter.socket_cache import DoHSocket, DoTSocket, PlainSocket, QUICSocket, SocketCache


@pytest.mark.parametrize(
    ("protocol", "server"),
    [
        ("udp", "dns.google"),
        ("tcp", "dns.google"),
        ("udptcp", "dns.google"),
        ("dot", "anycast.uncensoreddns.org"),
        ("doh", "anycast.uncensoreddns.org"),
        ("doq", "dns-unfiltered.adguard.com"),
        ("doh3", "dns-unfiltered.adguard.com"),
    ],
)
def test_socket_reuse_enabled(dns_exporter_example_config, protocol, server):
    """Test socket reuse enabled for all protocols."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": server,
            "family": "ipv4",
            "protocol": protocol,
            "connection_reuse": True,
        },
        timeout=5,
    )
    assert f'server="{protocol}://{server}:' in r.text
    assert "dnsexp_dns_query_success 1.0" in r.text
    assert "connection=" not in r.text


def test_socket_reuse_enabled_udptcp_large_reply(dns_exporter_example_config):
    """Test fallback to TCP with a large reply (dr.dk TXT is around 4kb)."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "dr.dk",
            "query_type": "txt",
            "server": "dns.google",
            "family": "ipv4",
            "protocol": "udptcp",
            "connection_reuse": True,
        },
        timeout=5,
    )
    assert 'transport="TCP"' in r.text


@pytest.mark.parametrize(
    ("protocol", "server"),
    [
        ("udp", "dns.google"),
        ("tcp", "dns.google"),
        ("udptcp", "dns.google"),
        ("dot", "anycast.uncensoreddns.org"),
        ("doh", "anycast.uncensoreddns.org"),
        ("doq", "dns-unfiltered.adguard.com"),
        ("doh3", "dns-unfiltered.adguard.com"),
    ],
)
def test_socket_reuse_disabled(dns_exporter_example_config, protocol, server):
    """Test socket reuse disabled for all protocols."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": server,
            "family": "ipv4",
            "protocol": protocol,
            "connection_reuse": False,
        },
        timeout=5,
    )
    assert f'server="{protocol}://{server}:' in r.text
    assert "dnsexp_dns_query_success 1.0" in r.text


@pytest.mark.usefixtures("dns_exporter_example_config_connection_label")
@pytest.mark.parametrize(
    ("protocol", "server"),
    [
        ("udp", "dns.google"),
        ("tcp", "dns.google"),
        ("udptcp", "dns.google"),
        ("dot", "anycast.uncensoreddns.org"),
        ("doh", "anycast.uncensoreddns.org"),
        ("doq", "dns-unfiltered.adguard.com"),
        ("doh3", "dns-unfiltered.adguard.com"),
    ],
)
def test_connection_label_reuse_false(protocol, server):
    """Make sure a connection label is added."""
    r = requests.get(
        "http://127.0.0.1:15355/query",
        params={
            "query_name": "example.com",
            "server": server,
            "family": "ipv4",
            "protocol": protocol,
            "connection_reuse": False,
        },
        timeout=5,
    )
    assert f'server="{protocol}://{server}:' in r.text
    assert "dnsexp_dns_query_success 1.0" in r.text
    assert "connection=" in r.text


@pytest.mark.usefixtures("dns_exporter_example_config_connection_label")
@pytest.mark.parametrize(
    ("protocol", "server"),
    [
        ("udp", "dns.google"),
        ("tcp", "dns.google"),
        ("udptcp", "dns.google"),
        ("dot", "anycast.uncensoreddns.org"),
        ("doh", "anycast.uncensoreddns.org"),
        ("doq", "dns-unfiltered.adguard.com"),
        ("doh3", "dns-unfiltered.adguard.com"),
    ],
)
def test_connection_label_reuse_true(protocol, server, dns_exporter_example_config_connection_label):
    """Make sure a connection label is added."""
    r = requests.get(
        "http://127.0.0.1:15355/query",
        params={
            "query_name": "example.com",
            "server": server,
            "family": "ipv4",
            "protocol": protocol,
            "connection_reuse": True,
        },
        timeout=5,
    )
    assert f'server="{protocol}://{server}:' in r.text
    assert "dnsexp_dns_query_success 1.0" in r.text
    assert "connection=" in r.text


def test_connection_reuse_retry(caplog, dns_exporter_example_config, mock_get_dns_response_tcp_eoferror):
    """Make sure a socket error is returned if a socket in the cache has been closed."""
    caplog.set_level(logging.DEBUG)
    requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "ip": "8.8.8.8",
            "family": "ipv4",
            "protocol": "tcp",
            "connection_reuse": True,
        },
        timeout=5,
    )
    assert "failure reason is 'socket_error'" in caplog.text


def test_no_connection_reuse_retry(caplog, dns_exporter_example_config, mock_get_dns_response_tcp_eoferror):
    """Make sure an exception is raised on EOFError when conn reuse is disabled."""
    caplog.set_level(logging.DEBUG)
    requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "ip": "8.8.8.8",
            "family": "ipv4",
            "protocol": "tcp",
            "connection_reuse": False,
        },
        timeout=5,
    )
    assert "failure reason is 'socket_error'" in caplog.text


def test_plain_socket_cache_delete(exporter, caplog):
    """Make sure deleting from the socket cache works for plaintext sockets."""
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="tcp",
            server="dns.google",
            query_name="example.com",
            ip="8.8.8.8",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    cachekey = socket_cache.get_cache_key(config=config)
    sock, reused = socket_cache.get_plaintext_socket(config=config)
    assert isinstance(sock, PlainSocket)
    assert isinstance(sock.socket, socket.socket)
    assert not reused
    assert len(socket_cache.plain_sockets[cachekey].sockets) == 1
    socket_cache.delete_socket(sock=sock)
    assert len(socket_cache.plain_sockets[cachekey].sockets) == 0


def test_dot_socket_cache_delete(exporter, caplog):
    """Make sure deleting from the socket cache works for dot sockets."""
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="dot",
            server="dns.google",
            query_name="example.com",
            ip="8.8.8.8",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    cachekey = socket_cache.get_cache_key(config=config)
    sock, reused = socket_cache.get_dot_socket(config=config, verify=True)
    assert isinstance(sock, DoTSocket)
    assert isinstance(sock.socket, ssl.SSLSocket)
    assert not reused
    assert len(socket_cache.dot_sockets[cachekey].sockets) == 1
    socket_cache.delete_socket(sock=sock)
    assert len(socket_cache.dot_sockets[cachekey].sockets) == 0


def test_doh_socket_cache_delete(exporter, caplog):
    """Make sure deleting from the socket cache works for doh2 sockets."""
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="doh",
            server="dns.google",
            query_name="example.com",
            ip="8.8.8.8",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    cachekey = socket_cache.get_cache_key(config=config)
    sock, reused = socket_cache.get_doh_socket(config=config, verify=True)
    assert isinstance(sock, DoHSocket)
    assert isinstance(sock.socket, httpx.Client)
    assert not reused
    assert len(socket_cache.doh_sockets[cachekey].sockets) == 1
    socket_cache.delete_socket(sock=sock)
    assert len(socket_cache.doh_sockets[cachekey].sockets) == 0


def test_quic_socket_cache_delete(exporter, caplog):
    """Make sure deleting from the socket cache works for quic sockets."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="doq",
            server="dns-unfiltered.adguard.com",
            query_name="example.com",
            ip="94.140.14.140",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    cachekey = socket_cache.get_cache_key(config=config)
    sock, reused = socket_cache.get_quic_socket(config=config, verify=True)
    assert isinstance(sock, QUICSocket)
    assert isinstance(sock.socket, SyncQuicConnection)
    assert not reused
    assert len(socket_cache.quic_sockets[cachekey].sockets) == 1, "expected 1 QUIC socket before deleting"
    # make sure detecting a closed connection works
    sock.socket._done = True  # noqa: SLF001
    sock, reused = socket_cache.get_quic_socket(config=config, verify=True)
    assert "Deleting stale QUIC socket" in caplog.text
    assert len(socket_cache.quic_sockets[cachekey].sockets) == 1
    socket_cache.delete_socket(sock=sock)
    assert len(socket_cache.quic_sockets[cachekey].sockets) == 0


def test_plain_socket_cache_key_equality(exporter, caplog):
    """Test comparing socket cache keys."""
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="tcp",
            server="dns.google",
            query_name="example.com",
            ip="8.8.8.8",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    key1 = socket_cache.get_cache_key(config=config)
    key2 = socket_cache.get_cache_key(config=config)
    assert key1 == key2, "__eq__ in SocketCacheKey doesn't eq"
    assert key1.__hash__() == key2.__hash__(), "__hash__ in SocketCacheKey doesn't hash"
    assert not key1.__eq__(config)
    config.ip = "8.8.4.4"
    key3 = socket_cache.get_cache_key(config=config)
    assert key1 != key3


def test_cache_key_labels(exporter):
    """Make sure the labels property works."""
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="tcp",
            server="dns.google",
            query_name="example.com",
            ip="8.8.8.8",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    key = socket_cache.get_cache_key(config=config)
    assert key.labels == (key.protocol, key.server, key.ip, key.verify, key.proxy)


def test_cache_metrics(exporter, caplog):
    """Try calling the cache metrics update method."""
    caplog.set_level(logging.DEBUG)
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="tcp",
            server="dns.google",
            query_name="example.com",
            ip="8.8.8.8",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    socket_cache.get_plaintext_socket(config=config)
    socket_cache.update_metrics()
    assert "Updating SocketCache metrics for 1 plain_sockets" in caplog.text


def tcp_query(qname) -> requests.get:
    """Do a tcp query."""
    return requests.get(
        "http://127.0.0.1:15355/query",
        params={
            "query_name": qname,
            "server": "dns.google",
            "family": "ipv4",
            "protocol": "tcp",
            "connection_reuse": True,
            "ip": "8.8.8.8",
        },
        timeout=5,
    )


def test_plain_socket_locking(dns_exporter_example_config_connection_label):
    """Test plain socket locking under parallel use."""
    qnames = ["example.com", "example.net", "example.org", "example.com", "example.net", "example.org"]
    with ThreadPoolExecutor(max_workers=3) as executor:
        results = list(executor.map(tcp_query, qnames))
    for r in results:
        assert "dnsexp_dns_query_success 1.0" in r.text


def dot_query(qname) -> requests.get:
    """Do a DoT query."""
    return requests.get(
        "http://127.0.0.1:15355/query",
        params={
            "query_name": qname,
            "server": "dns.google",
            "family": "ipv4",
            "protocol": "dot",
            "connection_reuse": True,
            "ip": "8.8.8.8",
        },
        timeout=5,
    )


def test_dot_socket_locking(dns_exporter_example_config_connection_label):
    """Test dot socket locking under parallel use."""
    qnames = ["example.com", "example.net", "example.org", "example.com", "example.net", "example.org"]
    with ThreadPoolExecutor(max_workers=3) as executor:
        results = list(executor.map(dot_query, qnames))
    for r in results:
        assert "dnsexp_dns_query_success 1.0" in r.text


def doh_query(qname) -> requests.get:
    """Do a DoH query."""
    return requests.get(
        "http://127.0.0.1:15355/query",
        params={
            "query_name": qname,
            "server": "dns.google",
            "family": "ipv4",
            "protocol": "doh",
            "connection_reuse": True,
            "ip": "8.8.8.8",
        },
        timeout=5,
    )


def test_doh_socket_locking(dns_exporter_example_config_connection_label):
    """Test doh socket locking under parallel use."""
    qnames = ["example.com", "example.net", "example.org", "example.com", "example.net", "example.org"]
    with ThreadPoolExecutor(max_workers=3) as executor:
        results = list(executor.map(doh_query, qnames))
    for r in results:
        assert "dnsexp_dns_query_success 1.0" in r.text


def doq_query(qname) -> requests.get:
    """Do a DoQ query."""
    return requests.get(
        "http://127.0.0.1:15355/query",
        params={
            "query_name": qname,
            "server": "dns-unfiltered.adguard.com",
            "family": "ipv4",
            "protocol": "doq",
            "connection_reuse": True,
            "ip": "94.140.14.141",
        },
        timeout=5,
    )


def test_doq_socket_locking(dns_exporter_example_config_connection_label):
    """Test doq socket locking under parallel use."""
    qnames = ["example.com", "example.net", "example.org", "example.com", "example.net", "example.org"]
    with ThreadPoolExecutor(max_workers=3) as executor:
        results = list(executor.map(doq_query, qnames))
    for r in results:
        assert "dnsexp_dns_query_success 1.0" in r.text


def test_plain_socket_cache_delete_oserror(exporter, caplog, mock_socket_close_oserror):
    """Make sure deleting from the plain socket cache handles OSErrors."""
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="tcp",
            server="dns.google",
            query_name="example.com",
            ip="8.8.8.8",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    cachekey = socket_cache.get_cache_key(config=config)
    sock, _ = socket_cache.get_plaintext_socket(config=config)
    assert isinstance(sock, PlainSocket)
    assert isinstance(sock.socket, socket.socket)
    assert len(socket_cache.plain_sockets[cachekey].sockets) == 1
    socket_cache.delete_socket(sock=sock)
    assert len(socket_cache.plain_sockets[cachekey].sockets) == 0


def test_dot_socket_cache_delete_oserror(exporter, caplog, mock_socket_close_oserror):
    """Make sure deleting from the dot socket cache handles OSErrors."""
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="dot",
            server="dns.google",
            query_name="example.com",
            ip="8.8.8.8",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    cachekey = socket_cache.get_cache_key(config=config)
    sock, _ = socket_cache.get_dot_socket(config=config, verify=True)
    assert isinstance(sock, DoTSocket)
    assert isinstance(sock.socket, ssl.SSLSocket)
    assert len(socket_cache.dot_sockets[cachekey].sockets) == 1
    socket_cache.delete_socket(sock=sock)
    assert len(socket_cache.dot_sockets[cachekey].sockets) == 0


@pytest.mark.skipif(
    sys.version_info[:2] == (3, 12),
    reason="Test is slow and weird on 3.12, skip for now https://github.com/tykling/dns_exporter/issues/203",
)
def test_socket_cache_cleanup_thread_age():
    """Make sure socket cache housekeeping works for old sockets."""
    with subprocess.Popen(
        args=[
            "dns_exporter",
            "-p",
            "15354",
            "--connection-cleanup-interval-seconds",
            "1",
            "--connection-max-age-seconds",
            "1",
            "--debug",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as proc:
        time.sleep(2)
        if proc.poll():
            # process didn't start properly, bail out
            pytest.fail(
                "Unable to create test instance on 127.0.0.1:15354",
            )
        ######################################
        requests.get(
            "http://127.0.0.1:15354/query",
            params={
                "query_name": "example.com",
                "server": "dns.google",
                "family": "ipv4",
                "ip": "8.8.8.8",
                "connection_reuse": "true",
            },
            timeout=5,
        )
        r = requests.get(
            "http://127.0.0.1:15354/metrics",
            timeout=5,
        )
        assert (
            'dnsexp_socket_uses_total{index="0",ip="8.8.8.8",protocol="udp",proxy="none",server="udp://dns.google:53",verify="none"} 1.0'
            in r.text
        )
        time.sleep(3)
        r = requests.get(
            "http://127.0.0.1:15354/metrics",
            timeout=5,
        )
        assert (
            'dnsexp_socket_uses_total{index="0",ip="8.8.8.8",protocol="udp",proxy="none",server="udp://dns.google:53",verify="none"} 1.0'
            not in r.text
        )
        ######################################
        proc.terminate()
        output = proc.stderr.read().decode()
        assert (
            "Deleting socket SocketCacheKey(protocol='udp', server='udp://dns.google:53', ip='8.8.8.8', verify='none', proxy='none') index 0 due to age"
            in output
        )


@pytest.mark.skipif(
    sys.version_info[:2] == (3, 12),
    reason="Test is slow and weird on 3.12, skip for now https://github.com/tykling/dns_exporter/issues/203",
)
def test_socket_cache_cleanup_thread_idle():
    """Make sure socket cache housekeeping works for idle sockets."""
    with subprocess.Popen(
        args=[
            "dns_exporter",
            "-p",
            "15354",
            "--connection-cleanup-interval-seconds",
            "1",
            "--connection-max-idle-seconds",
            "1",
            "--debug",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as proc:
        time.sleep(2)
        if proc.poll():
            # process didn't start properly, bail out
            pytest.fail(
                "Unable to create test instance on 127.0.0.1:15354",
            )
        ######################################
        requests.get(
            "http://127.0.0.1:15354/query",
            params={
                "query_name": "example.com",
                "server": "dns.google",
                "family": "ipv4",
                "ip": "8.8.8.8",
                "connection_reuse": "true",
            },
            timeout=5,
        )
        r = requests.get(
            "http://127.0.0.1:15354/metrics",
            timeout=5,
        )
        assert (
            'dnsexp_socket_uses_total{index="0",ip="8.8.8.8",protocol="udp",proxy="none",server="udp://dns.google:53",verify="none"} 1.0'
            in r.text
        )
        time.sleep(3)
        r = requests.get(
            "http://127.0.0.1:15354/metrics",
            timeout=5,
        )
        assert (
            'dnsexp_socket_uses_total{index="0",ip="8.8.8.8",protocol="udp",proxy="none",server="udp://dns.google:53",verify="none"} 1.0'
            not in r.text
        )
        ######################################
        proc.terminate()
        output = proc.stderr.read().decode()
        assert (
            "Deleting socket SocketCacheKey(protocol='udp', server='udp://dns.google:53', ip='8.8.8.8', verify='none', proxy='none') index 0 with idle time"
            in output
        )


def test_socket_delete_retry(exporter, dns_exporter_example_config, mock_get_dns_response_tcp_eoferror, caplog):
    """Make sure a socket is deleted if it raises an error."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    socket_cache = SocketCache()
    prepared = exporter.prepare_config(
        ConfigDict(
            protocol="tcp",
            server="dns.google",
            query_name="example.com",
            ip="8.8.8.8",
            family="ipv4",
        )
    )
    config = Config.create(name="test", **prepared)
    sock, _ = socket_cache.get_plaintext_socket(config=config)
    # make sure detecting a closed connection works
    sock.socket.shutdown(socket.SHUT_RDWR)
    sock.socket.close()
    requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "family": "ipv4",
            "protocol": "tcp",
            "ip": "8.8.8.8",
            "connection_reuse": True,
        },
        timeout=5,
    )
    assert "failure reason is 'socket_error'" in caplog.text



"""tests for example configuration snippets."""

import shutil
import time

import pytest
import requests

# the tests in this module require Prometheus to be installed and in $PATH
prom = shutil.which("prometheus")
if prom is None:
    pytest.skip(
        "Skipping example tests because Prometheus is not installed",
        allow_module_level=True,
    )


@pytest.mark.parametrize(
    "prometheus_server",
    ["list_of_servers/prometheus.yml"],
    indirect=True,
)
@pytest.mark.parametrize(
    "dns_exporter_param_config",
    ["list_of_servers/dns_exporter.yml"],
    indirect=True,
)
def test_list_of_servers(prometheus_server, dns_exporter_param_config):
    """Test the list_of_servers snippets from the docs."""
    for _ in range(15):
        r = requests.get(
            "http://127.0.0.1:9092/api/v1/query?query=dnsexp_dns_query_success==1",
            timeout=5,
        )
        if len(r.json()["data"]["result"]) > 0 and r.json()["data"]["result"][0]["value"][1] == "1":
            break
        time.sleep(1)
    else:
        pytest.fail("expected result not found in prom")


@pytest.mark.parametrize(
    "prometheus_server",
    ["list_of_names/prometheus.yml"],
    indirect=True,
)
@pytest.mark.parametrize(
    "dns_exporter_param_config",
    ["list_of_names/dns_exporter.yml"],
    indirect=True,
)
def test_list_of_names(caplog, prometheus_server, dns_exporter_param_config):
    """Test the list_of_names snippets from the docs."""
    for _ in range(15):
        r = requests.get(
            "http://127.0.0.1:9092/api/v1/query?query=dnsexp_dns_query_success==1",
            timeout=5,
        )
        if len(r.json()["data"]["result"]) > 0 and r.json()["data"]["result"][0]["value"][1] == "1":
            break
        time.sleep(1)
    else:
        pytest.fail("expected result not found in prom")



"""``dns_exporter.entrypoint`` contains argparse stuff and ``dns_exporter`` script entrypoint.

This module is mostly boilerplate code for command-line argument handling and logging.
"""

from __future__ import annotations

import argparse
import logging
import os
import signal
import sys
import threading
import warnings
from http.server import ThreadingHTTPServer
from pathlib import Path
from typing import TYPE_CHECKING

import yaml

from dns_exporter.config import ConfigDict
from dns_exporter.exceptions import CleanupAndExit
from dns_exporter.exporter import DNSExporter
from dns_exporter.socket_cache import SocketCache, cleanup_socket_cache

if TYPE_CHECKING:
    from types import FrameType

# get logger
logger = logging.getLogger(f"dns_exporter.{__name__}")


def get_parser() -> argparse.ArgumentParser:
    """Create and return the argparse object."""
    parser = argparse.ArgumentParser(
        description=f"dns_exporter version {DNSExporter.__version__}. See ReadTheDocs for more info.",
    )

    # optional arguments
    parser.add_argument(
        "-c",
        "--config-file",
        dest="config-file",
        help="The path to the yaml config file to use. Only the root 'modules' key is read from the config file.",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_const",
        dest="log-level",
        const="DEBUG",
        help="Debug mode. Equal to setting --log-level=DEBUG.",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "-L",
        "--listen-ip",
        type=str,
        help="Listen IP. Defaults to 127.0.0.1. Set to :: to listen on all v6 IPs.",
        default="127.0.0.1",
    )
    parser.add_argument(
        "-l",
        "--log-level",
        dest="log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Logging level. One of DEBUG, INFO, WARNING, ERROR, CRITICAL. Defaults to INFO.",
        default="INFO",
    )
    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        type=int,
        help="The port the exporter should listen for requests on. Default: 15353",
        default=15353,
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_const",
        dest="log-level",
        const="WARNING",
        help="Quiet mode. No output at all if no errors are encountered. Equal to setting --log-level=WARNING.",
        default=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--connection-max-age-seconds",
        type=int,
        help=(
            "The maximum age in seconds for connection reuse entries. "
            "0 means never destroy connections due to age. Default: 0"
        ),
        default=0,
    )
    parser.add_argument(
        "--connection-max-idle-seconds",
        type=int,
        help=(
            "The maximum idle time in seconds for connection reuse entries. "
            "0 means never destroy connections due to idle time. Default: 3600"
        ),
        default=3600,
    )
    parser.add_argument(
        "--connection-cleanup-interval-seconds",
        type=int,
        help=(
            "The interval in seconds between connection reuse housekeeping. "
            "Set to 0 to disable socket cache housekeeping entirely. Default: 600"
        ),
        default=600,
    )
    parser.add_argument(
        "-v",
        "--version",
        dest="version",
        action="store_true",
        help="Show version and exit.",
        default=argparse.SUPPRESS,
    )
    return parser


def parse_args(
    mockargs: list[str] | None = None,
) -> tuple[argparse.ArgumentParser, argparse.Namespace]:
    """Create an argparse monster and parse mockargs or sys.argv[1:]."""
    parser = get_parser()
    args = parser.parse_args(mockargs or sys.argv[1:])
    return parser, args


def configure_logging(args: argparse.Namespace) -> None:
    """Configure the log format and level."""
    console_logformat = "%(asctime)s %(levelname)s %(name)s.%(funcName)s():%(lineno)i:  %(message)s"
    level = getattr(args, "log-level")
    logging.basicConfig(
        level=level,
        format=console_logformat,
        datefmt="%Y-%m-%d %H:%M:%S %z",
    )
    logger.setLevel(level)
    # also configure the root logger
    rootlogger = logging.getLogger("")
    rootlogger.setLevel(level)
    # httpx is noisy at INFO
    if level == "INFO":
        # httpx is noisy at level info, cap to WARNING
        logging.getLogger("httpx").setLevel(logging.WARNING)
    logger.info(
        f"dns_exporter v{DNSExporter.__version__} starting up - logging at level {level}",
    )
    if os.getenv("DNSEXP_CONNECTION_LABEL"):
        logger.info("DNSEXP_CONNECTION_LABEL set - enabling 'connection' label feature")  # pragma: no cover
    else:
        logger.info("DNSEXP_CONNECTION_LABEL unset - disabling 'connection' label feature")


def initialise_socket_cache(args: argparse.Namespace) -> tuple[SocketCache, threading.Thread | None]:
    """Initialise socket cache and socket cache housekeeping thread."""
    socket_cache = SocketCache()
    # configure socket cache
    socket_cache.socket_max_age_seconds = args.connection_max_age_seconds
    socket_cache.socket_max_idle_seconds = args.connection_max_idle_seconds
    socket_cache.housekeeping_interval = args.connection_cleanup_interval_seconds
    socket_cache.housekeeping_exit_event = threading.Event()
    logger.debug(
        f"SocketCache initialised with max. age {socket_cache.socket_max_age_seconds} "
        f"seconds and max. idle {socket_cache.socket_max_idle_seconds} seconds "
        f"and housekeeping interval {socket_cache.housekeeping_interval} seconds"
    )

    if args.connection_cleanup_interval_seconds > 0:
        # start socket housekeeping background thread
        housekeeping_thread = threading.Thread(target=socket_cache.housekeeping, args=())
        housekeeping_thread.daemon = True
        housekeeping_thread.start()
        logger.debug(f"Started socket housekeeping background thread {housekeeping_thread}")
    else:
        logger.debug("Not starting housekeeping thread")
        housekeeping_thread = None
    return socket_cache, housekeeping_thread


def main(mockargs: list[str] | None = None) -> None:
    """Read config and start exporter."""
    # suppress warnings at runtime
    if not sys.warnoptions:
        warnings.simplefilter("ignore")

    # get arpparser and parse args
    _, args = parse_args(mockargs)

    # handle version check
    if hasattr(args, "version"):
        print(f"dns_exporter version {DNSExporter.__version__}")  # noqa: T201
        sys.exit(0)

    # configure logging
    configure_logging(args=args)
    logger.debug(f"dns_exporter parsed command-line arguments: {mockargs or sys.argv[1:]}")

    if hasattr(args, "config-file"):
        with Path(getattr(args, "config-file")).open() as f:
            try:
                configfile = yaml.load(f, Loader=yaml.SafeLoader)
            except Exception:
                logger.exception(
                    f"Unable to parse YAML config file {getattr(args, 'config-file')} - bailing out.",
                )
                sys.exit(1)
        if (
            not configfile
            or "modules" not in configfile
            or not isinstance(configfile["modules"], dict)
            or not configfile["modules"]
        ):
            # configfile is empty, missing "modules" key, or modules is empty or not a dict
            logger.error(
                f"Invalid config file {getattr(args, 'config-file')} - yaml was valid but no modules found",
            )
            sys.exit(1)
        logger.debug(f"Read {len(configfile['modules'])} modules from config file {getattr(args, 'config-file')}:")
        logger.debug(list(configfile["modules"].keys()))
    else:
        # there is no config file
        configfile = {"modules": {}}
        logger.debug(
            "No -c / --config-file found so a config file will not be used. No modules loaded.",
        )

    # initialise the socket cache and housekeeping thread
    socket_cache, housekeeping_thread = initialise_socket_cache(args=args)

    # configure DNSExporter handler and start HTTPServer
    handler = DNSExporter
    if configfile["modules"] and not handler.configure(
        modules={k: ConfigDict(**v) for k, v in configfile["modules"].items()},  # type: ignore[typeddict-item]
    ):
        logger.error(
            "An error occurred while configuring dns_exporter. Bailing out.",
        )
        sys.exit(1)

    # Usually main() runs in the main Python thread. Skip configuring signal handler if it does not.
    if threading.current_thread() is threading.main_thread():
        logger.debug("Running in main thread, connecting signal handlers...")
        # this is the main thread, it is safe to do signal handling
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
    else:
        logger.warning("Not running in main thread, skipping signal handlers...")

    logger.info(
        f"Ready to serve requests. Starting listener on {args.listen_ip} port {args.port}...",
    )
    try:
        ThreadingHTTPServer((args.listen_ip, args.port), handler).serve_forever()
    except OSError:
        logger.exception(
            f"Unable to start listener, maybe port {args.port} is in use? bailing out",
        )
        sys.exit(1)
    except CleanupAndExit:
        logger.info("Signal received, cleaning up before exit...")
    finally:
        cleanup_socket_cache(socket_cache=socket_cache, housekeeping_thread=housekeeping_thread)
        logger.info("Clean exit - goodbye for now :)")


def signal_handler(sig: int, frame: FrameType | None) -> None:
    """This signal handler raises KeyboardInterrupt to allow cleanup before exit."""
    logger.debug(f"Signal {sig} received in frame {frame}, raising CleanupAndExit to trigger cleanup and exit...")
    raise CleanupAndExit


if __name__ == "__main__":  # pragma: no cover
    main()



"""dnsexporter tests for the exporter module."""

import logging

import dns.flags
import pytest
import requests

from dns_exporter.config import Config, ConfigDict, RFValidator, RRValidator
from dns_exporter.entrypoint import main
from dns_exporter.exporter import get_query


def test_main_no_config(dns_exporter_main_no_config_no_debug):
    """Test basic functionality on a debug-less instance with no modules loaded."""
    r = requests.get(
        "http://127.0.0.1:35353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert r.status_code == 200, "non-200 returncode"
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_noconfig_server(dns_exporter_no_main_no_config):
    """Test basic lookup functionality on an instance with debug enabled and no modules loaded."""
    r = requests.get(
        "http://127.0.0.1:45353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_config_server(dns_exporter_example_config):
    """Test basic lookup functionality on an instance with modules loaded."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_config_endpoint(dns_exporter_example_config):
    """Test the /config endpoint."""
    r = requests.get(
        "http://127.0.0.1:25353/config",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "protocol": "tcp",
            "proxy": "socks5://127.0.0.1:1081",
        },
        timeout=5,
    )
    config = r.json()
    assert config["server"] == "tcp://dns.google:53"
    assert config["query_name"] == "example.com"


def test_config_endpoint_2(dns_exporter_example_config):
    """Test the /config endpoint some more."""
    r = requests.get(
        "http://127.0.0.1:25353/config",
        params={
            "module": "cf_doh",
        },
        timeout=5,
    )
    config = r.json()
    assert config["protocol"] == "doh"
    assert config["server"] == "doh://1dot1dot1dot1.cloudflare-dns.com:443/dns-query"
    assert config["query_name"] == "bornhack.dk"
    assert config["query_type"] == "NS"
    assert config["validate_response_flags"]["fail_if_any_absent"] == ["AD"]


def test_invalid_qs_ip(dns_exporter_example_config):
    """Trigger an invalid_request_ip failure."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "ip": "notanip",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_invalid_configfile_ip(caplog, exporter):
    """Make sure a message is logged when parsing an IP fails."""
    caplog.clear()
    exporter.configure(modules={"test": {"ip": "notanip"}})
    assert "Unable to parse IP address notanip" in caplog.text


def test_missing_query_name(dns_exporter_example_config):
    """Trigger an invalid_request_query_name failure."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_missing_server(dns_exporter_example_config):
    """Trigger an invalid_request_server failure."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_undefined_module(dns_exporter_example_config, caplog):
    """Trigger an invalid_request_module failure."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "module": "notamodule",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_unknown_config_key(dns_exporter_example_config, caplog):
    """Trigger an invalid_request_config failure by providing an unknown key."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "foo": "bar",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_ip_family_conflict(dns_exporter_example_config, caplog):
    """Trigger an invalid_request_ip failure by providing a v4 ip and v6 family."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv6",
            "ip": "192.0.2.53",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_ip_conflict(dns_exporter_example_config, caplog):
    """Trigger an invalid_request_ip failure by providing a server ip and a conflicting ip."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "192.0.2.1",
            "query_name": "example.com",
            "ip": "192.0.2.53",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_ip_and_hostname(dns_exporter_example_config, caplog):
    """Make sure using an IP and hostname results in the IP overriding the server hostname DNS lookup."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "ip": "8.8.4.4",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert "Using server IP 8.8.4.4 (from config) for the DNS server connection" in caplog.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_unresolvable_server(dns_exporter_example_config, caplog):
    """Trigger an invalid_request_server failure by providing an unresolvable server hostname."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "notaserver.example",
            "query_name": "example.com",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_ipv6_family(dns_exporter_example_config, caplog):
    """Test IP override for ipv6 family."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "ip": "2001:4860:4860::8888",
            "timeout": 0.1,
        },
        timeout=5,
    )
    assert "Using server IP 2001:4860:4860::8888 (from config) for the DNS server connection" in caplog.text


def test_ipv7_family(dns_exporter_example_config, caplog):
    """Trigger an invalid_request_family failure with ipv7."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv7",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_index(dns_exporter_example_config, caplog):
    """Test the index page served at /."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/",
        timeout=5,
    )
    assert r.status_code == 200, "non-200 returncode"
    assert "DNS Exporter" in r.text
    assert "Returning index page for request to /" in caplog.text


def test_404(dns_exporter_example_config, caplog):
    """Make sure a 404 is returned from unknown endpoints."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/notfound",
        timeout=5,
    )
    assert r.status_code == 404, "non-404 returncode"
    assert "404 not found" in r.text
    assert "Unknown endpoint '/notfound' returning 404" in caplog.text


def test_tcp(dns_exporter_example_config, caplog):
    """Test basic tcp protocol functionality."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "protocol": "tcp",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert 'transport="TCP"' in r.text
    assert "Protocol tcp got a DNS query response over TCP" in caplog.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_udptcp(dns_exporter_example_config, caplog):
    """Test basic udptcp functionality."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "protocol": "udptcp",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert 'transport="UDP"' in r.text
    assert "Protocol udptcp got a DNS query response over UDP" in caplog.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_udptcp_fallback(dns_exporter_example_config, caplog):
    """Test udptcp fallback to TCP functionality."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "dr.dk",
            "protocol": "udptcp",
            "family": "ipv4",
            "query_type": "TXT",
        },
        timeout=5,
    )
    assert 'transport="TCP"' in r.text
    assert "Protocol udptcp got a DNS query response over TCP" in caplog.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_dot(dns_exporter_example_config, caplog):
    """Test basic DoT functionality."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "protocol": "dot",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert 'transport="TCP"' in r.text
    assert "Protocol dot got a DNS query response over TCP" in caplog.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_doh(dns_exporter_example_config, caplog):
    """Test basic DoH functionality."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "protocol": "doh",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert 'transport="TCP"' in r.text
    assert 'protocol="doh"' in r.text
    assert "Protocol doh got a DNS query response over TCP" in caplog.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_doh_httpx_writeerror(dns_exporter_example_config, caplog, mock_dns_query_https_httx_writeerror):
    """Test DoH error handling for httpx.WriteError."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "protocol": "doh",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert "Protocol doh raised exception, returning connection_error" in caplog.text
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_doh3(dns_exporter_example_config, caplog):
    """Test basic DoH3 functionality."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns-unfiltered.adguard.com",
            "query_name": "example.com",
            "protocol": "doh3",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert 'transport="QUIC"' in r.text
    assert 'protocol="doh3"' in r.text
    assert "Protocol doh3 got a DNS query response over QUIC" in caplog.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_doq(dns_exporter_example_config, caplog, recwarn):
    """Test basic DoQ functionality."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns-unfiltered.adguard.com",
            "query_name": "example.com",
            "protocol": "doq",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert 'transport="QUIC"' in r.text
    assert 'protocol="doq"' in r.text
    assert "Protocol doq got a DNS query response over QUIC" in caplog.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_edns_pad(dns_exporter_example_config, caplog):
    """Test edns_pad config setting."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv4",
            "edns_pad": 20,
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_no_edns(dns_exporter_example_config, caplog):
    """Test disabling EDNS."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv4",
            "ip": "8.8.4.4",
            "edns": False,
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text
    assert 'nsid="no_nsid"' in r.text


def test_broken_yaml_config(caplog, dns_exporter_broken_yaml_configfile):
    """Test loading a broken yaml config file."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    with pytest.raises(SystemExit) as e:
        main(["-c", str(dns_exporter_broken_yaml_configfile)])
    assert "Unable to parse YAML config file" in caplog.text
    assert e.value.code == 1, "Exit code not 1 as expected with broken yaml config"


def test_empty_yaml_config(caplog, dns_exporter_empty_yaml_configfile):
    """Test loading an empty yaml config file."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    with pytest.raises(SystemExit) as e:
        main(["-c", str(dns_exporter_empty_yaml_configfile)])
    assert "Invalid config file" in caplog.text
    assert e.value.code == 1, "Exit code not 1 as expected with empty yaml config"


def test_invalid_yaml_config(caplog, dns_exporter_invalid_yaml_configfile):
    """Test loading a yaml config file with invalid config options in it."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    with pytest.raises(SystemExit) as e:
        main(["-c", str(dns_exporter_invalid_yaml_configfile)])
    assert "An error occurred while configuring dns_exporter" in caplog.text
    assert e.value.code == 1, "Exit code not 1 as expected with invalid yaml config"


def test_configure(caplog, exporter):
    """Test calling configure() on the DNSExporter class."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    exporter.modules = {}
    exporter.configure(modules={"test": {"ip": "127.0.0.1"}})
    assert len(exporter.modules) == 1
    assert "1 module(s) loaded OK, total modules: 1." in caplog.text


def test_invalid_integer(dns_exporter_example_config, caplog):
    """Test passing an invalid integer to an integer setting."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "edns_bufsize": "foo",
        },
        timeout=5,
    )
    assert "Unable to validate integer for key edns_bufsize" in caplog.text
    assert "ValueError: invalid literal for int() with base 10: 'foo'" in caplog.text
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_configure_rrvalidator(caplog, exporter):
    """Test loading a config which creates an RRValidator object."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    exporter.modules = {}
    exporter.configure(
        modules={
            "test": {"validate_answer_rrs": RRValidator.create({"fail_if_count_eq": 4})},
        },
    )
    assert len(exporter.modules) == 1
    assert "1 module(s) loaded OK, total modules: 1." in caplog.text


def test_configure_rfvalidator(caplog, exporter):
    """Test loading a config which creates an RFValidator object."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    exporter.modules = {}
    exporter.configure(
        modules={
            "test": {
                "validate_response_flags": RFValidator.create(
                    {"fail_if_any_absent": ["peace", "love"]},
                ),
            },
        },
    )
    assert len(exporter.modules) == 1
    assert "1 module(s) loaded OK, total modules: 1." in caplog.text


def test_configure_bad_module(caplog, exporter):
    """Test the configure() method with an invalid module."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    exporter.modules = {}
    exporter.configure(modules={"test": {"query_class": "OUT"}})
    assert len(exporter.modules) == 0
    assert "Invalid value found while building config {'query_class': 'OUT'}" in caplog.text


def test_catch_unknown_exception(
    dns_exporter_example_config,
    mock_collect_zerodivisionerror,
):
    """Trigger an other_failure failure by raising a ValueError from get_dns_response()."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "query_name": "example.com",
            "server": "dns.google",
            "ip": "8.8.8.8",
            "family": "ipv4",
            "protocol": "doh",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


### ttl tests
def test_ttl(dns_exporter_example_config, caplog):
    """Test collecting ttl metrics."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "anycast.censurfridns.dk",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert 'rr_value="91.239.100.100"' in r.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_no_ttl(dns_exporter_no_main_no_config, caplog):
    """Test not collecting ttl metrics."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:45353/query",
        params={
            "server": "dns.google",
            "query_name": "anycast.censurfridns.dk",
            "family": "ipv4",
            "collect_ttl": False,
        },
        timeout=5,
    )
    assert 'rr_value="91.239.100.100"' not in r.text
    assert "dnsexp_dns_query_success 1.0" in r.text


### connection refused


@pytest.mark.parametrize("protocol", ["udp", "tcp", "udptcp", "dot", "doh", "doq"])
def test_connection_error_server(dns_exporter_example_config, caplog, protocol):
    """Trigger a connection_error failure for each protocol by connecting to something not listening."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "192.0.2.42:420",
            "protocol": protocol,
            "query_name": "example.org",
            "timeout": 1,
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


@pytest.mark.parametrize("protocol", ["udp", "tcp", "udptcp", "dot", "doh", "doq"])
def test_timeout_server(dns_exporter_example_config, caplog, protocol):
    """Trigger a timeout failure for each protocol by connecting to something not reachable."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "192.0.2.42:420",
            "protocol": protocol,
            "query_name": "example.org",
            "timeout": 0.00001,
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_collect_ttl_value_length(dns_exporter_example_config, caplog):
    """Test the collect_ttl_value_length setting."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.quad9.net",
            "query_name": "dns.google",
            "family": "ipv4",
            "collect_ttl_rr_value_length": 3,
        },
        timeout=5,
    )
    assert 'rr_value="8.8"' in r.text
    assert "dnsexp_dns_query_success 1.0" in r.text


def test_doh_bad_statuscode(dns_exporter_example_config, mock_dns_query_https_valuerror, caplog):
    """Test DoH fail with a bad statuscode."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "ip": "8.8.8.8",
            "query_name": "example.com",
            "protocol": "doh",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text
    assert "failure reason is 'invalid_response_statuscode'" in caplog.text


def test_doh_timeout(dns_exporter_example_config, mock_dns_query_httpx_connecttimeout, caplog):
    """Test DoH fail with a timeout."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "protocol": "doh",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text
    assert "failure reason is 'timeout'" in caplog.text


def test_nsid(dns_exporter_example_config, caplog):
    """Test edns nsid functionality."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "unicast.censurfridns.dk",
            "query_name": "example.com",
            "protocol": "dot",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert 'transport="TCP"' in r.text
    assert "Protocol dot got a DNS query response over TCP" in caplog.text
    assert "dnsexp_dns_query_success 1.0" in r.text
    assert 'nsid="unicast2.servers.censurfridns.dk"' in r.text


def test_query_rd_true(caplog, exporter):
    """Make sure the RD flag is set in the query when recursion_desired is true.

    Regression test for https://github.com/tykling/dns_exporter/issues/177
    """
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    prepared = exporter.prepare_config(
        ConfigDict(recursion_desired=True, query_name="example.com", server="dns.google")
    )
    config = Config.create(name="test", **prepared)
    q = get_query(config=config)
    assert dns.flags.RD in q.flags, "RD flag should be set"


def test_query_rd_false(caplog, exporter):
    """Make sure the RD flag is not set in the query when recursion_desired is false.

    Regression test for https://github.com/tykling/dns_exporter/issues/177
    """
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    prepared = exporter.prepare_config(
        ConfigDict(recursion_desired=False, query_name="example.com", server="dns.google")
    )
    config = Config.create(name="test", **prepared)
    q = get_query(config=config)
    assert dns.flags.RD not in q.flags, "RD flag should not be set"


def test_udp_truncated(dns_exporter_example_config, caplog):
    """Test UDP message truncation handling."""
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "dr.dk",
            "query_type": "TXT",
            "protocol": "udp",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert (
        "No DNS response received from server udp://dns.google:53 - failure reason is 'response_truncated'..."
        in caplog.text
    )
    assert "dnsexp_dns_query_success 0.0" in r.text



"""dnsexporter test of internal metrics. Runs last, breaks easily."""

import logging

import pytest
import requests

from dns_exporter.version import __version__


# run this test last
@pytest.mark.order(-1)
@pytest.mark.xfail(
    reason="This test is expected to fail whenever any test is rerun (which sometimes happen when testing with real-world servers)"
)
def test_internal_metrics(dns_exporter_example_config, tmp_path, caplog):
    """Test the internal metrics and make sure aggregated counts of all unit tests are there.

    This test breaks as soon as any other test breaks, it must run last, and it cannot run alone.
    """
    caplog.clear()
    caplog.set_level(logging.DEBUG)
    r = requests.get(
        "http://127.0.0.1:25353/metrics",
        timeout=5,
    )
    # write metrics to file to help debugging failing tests
    p = tmp_path / "metrics.prom"
    with p.open("w") as f:
        f.write(r.text)
    # on with the asserts
    assert f'dnsexp_build_version_info{{version="{__version__}"}} 1.0' in r.text
    assert "Returning exporter metrics for request to /metrics" in caplog.text
    for metric in [
        'dnsexp_http_requests_total{path="/notfound"} 1.0',
        'dnsexp_http_requests_total{path="/query"} 119.0',
        'dnsexp_http_requests_total{path="/config"} 4.0',
        'dnsexp_http_requests_total{path="/"} 1.0',
        'dnsexp_http_responses_total{path="/notfound",response_code="404"} 1.0',
        'dnsexp_http_responses_total{path="/query",response_code="200"} 119.0',
        'dnsexp_http_responses_total{path="/",response_code="200"} 1.0',
        "dnsexp_dns_queries_total 100.0",
        'dnsexp_dns_responsetime_seconds_bucket{additional="0",answer="6",authority="0",family="ipv4",flags="QR RA RD",ip="8.8.4.4",le="2.5",nsid="no_nsid",opcode="QUERY",port="53",protocol="udp",proxy="none",query_name="example.com",query_type="A",rcode="NOERROR",server="udp://dns.google:53",transport="UDP"}',
        'dnsexp_scrape_failures_total{additional="none",answer="none",authority="none",family="none",flags="none",ip="none",nsid="none",opcode="none",port="none",protocol="none",proxy="none",query_name="none",query_type="none",rcode="none",reason="invalid_request_config",server="none",transport="none"} 4.0',
        'dnsexp_scrape_failures_total{additional="none",answer="none",authority="none",family="none",flags="none",ip="none",nsid="none",opcode="none",port="none",protocol="none",proxy="none",query_name="none",query_type="none",rcode="none",reason="invalid_request_server",server="none",transport="none"} 2.0',
        'dnsexp_scrape_failures_total{additional="none",answer="none",authority="none",family="none",flags="none",ip="none",nsid="none",opcode="none",port="none",protocol="none",proxy="none",query_name="none",query_type="none",rcode="none",reason="invalid_request_query_name",server="none",transport="none"} 1.0',
        'dnsexp_scrape_failures_total{additional="none",answer="none",authority="none",family="none",flags="none",ip="none",nsid="none",opcode="none",port="none",protocol="none",proxy="none",query_name="none",query_type="none",rcode="none",reason="invalid_request_ip",server="none",transport="none"} 3.0',
        'dnsexp_scrape_failures_total{additional="none",answer="none",authority="none",family="ipv4",flags="none",ip="8.8.8.8",nsid="none",opcode="none",port="443",protocol="doh",proxy="none",query_name="example.com",query_type="A",rcode="none",reason="invalid_response_statuscode",server="doh://dns.google:443/dns-query",transport="none"} 1.0',
        'dnsexp_scrape_failures_total{additional="none",answer="none",authority="none",family="none",flags="none",ip="none",nsid="none",opcode="none",port="none",protocol="none",proxy="none",query_name="none",query_type="none",rcode="none",reason="invalid_request_module",server="none",transport="none"} 1.0',
        'dnsexp_scrape_failures_total{additional="none",answer="none",authority="none",family="none",flags="none",ip="none",nsid="none",opcode="none",port="none",protocol="none",proxy="none",query_name="none",query_type="none",rcode="none",reason="invalid_request_family",server="none",transport="none"} 1.0',
        'dnsexp_scrape_failures_total{additional="none",answer="none",authority="none",family="ipv4",flags="none",ip="8.8.8.8",nsid="none",opcode="none",port="443",protocol="doh",proxy="none",query_name="example.com",query_type="A",rcode="none",reason="other_failure",server="doh://dns.google:443/dns-query",transport="none"} 1.0',
        'dnsexp_scrape_failures_total{additional="none",answer="none",authority="none",family="none",flags="none",ip="none",nsid="none",opcode="none",port="none",protocol="none",proxy="none",query_name="none",query_type="none",rcode="none",reason="invalid_request_proxy",server="none",transport="none"} 2.0',
        'dnsexp_http_requests_total{path="/metrics"} 1.0',
    ]:
        assert metric in r.text, f"expected metric {metric} not found in metrics: {p}"



"""This module contains the custom exceptions used in dns_exporter."""


class ConfigError(Exception):
    """Exception class used when invalid config values are encountered."""


class ValidationError(Exception):
    """Exception class used when response validation fails."""

    def __init__(self, validator: str, reason: str) -> None:
        """Take validator and reason as arguments and raise the exception."""
        super().__init__(f"Response validator {validator} failed with reason {reason}", reason)


class UnknownFailureReasonError(RuntimeError):
    """Exception raised if an unknown failure reason is used (this is always a bug)."""

    def __init__(self, failure_reason: str) -> None:
        """Raise with failure reason."""
        super().__init__(f"Unknown failure_reason {failure_reason} - please file a bug!")


class ProtocolSpecificError(Exception):
    """Exception class used when DNS lookup fails with a protocol specific exception."""


class ConfigTypeError(TypeError):
    """Exception used when DNSCollector is initialised with a config that is not a dns_exporter.config.Config object."""


class QueryTypeError(TypeError):
    """Exception used when DNSCollector is initialised with a query that is not a dns.query.QueryMessage object."""


class LabelsTypeError(TypeError):
    """Exception used when DNSCollector is initialised with a labels arg that is not a list."""


class CleanupAndExit(Exception):  # noqa: N818
    """Exception raised to ask main loop to clean up and exit."""



"""dnsexporter tests for validator code."""

import logging
from contextlib import nullcontext as does_not_raise

import pytest
import requests
from dns.message import from_text

from dns_exporter.collector import DNSCollector
from dns_exporter.config import Config, RRValidator
from dns_exporter.exceptions import ValidationError

test_dns_response = from_text("""id 0
opcode QUERY
rcode NOERROR
flags QR RD RA
;QUESTION
www.example.com. IN A
;ANSWER
www.example.com. 1800 IN CNAME www.example.com.edgekey.net.
www.example.com.edgekey.net. 18416 IN CNAME www.example.com.edgekey.net.globalredir.exampledns.net.
www.example.com.edgekey.net.globalredir.exampledns.net. 900 IN CNAME abc.dscc.exampleedge.net.
abc.dscc.exampleedge.net. 20 IN A 10.8.23.42
abc.dscc.exampleedge.net. 20 IN AAAA fd00:dead:beef::23:42
;AUTHORITY
;ADDITIONAL
""")


def test_validate_rcode(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_rcode error by asking for an NXDOMAIN."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "donoevil.google.com",
            "family": "ipv4",
        },
        timeout=5,
    )
    assert 'rcode="NXDOMAIN"' in r.text
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_rcode_2(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_rcode error by asking for an NXDOMAIN."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "donoevil.google.com",
            "family": "ipv4",
            "valid_rcodes": "NXDOMAIN",
        },
        timeout=5,
    )
    assert 'rcode="NXDOMAIN"' in r.text
    assert "dnsexp_dns_query_success 1.0" in r.text


# validate_flags fail_if_any_absent


def test_validate_flags_fail_if_any_absent(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_flags failure by checking for AD flag on google.com."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "google.com",
            "family": "ipv4",
            "module": "has_ad",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_flags_fail_if_any_absent_2(dns_exporter_example_config, caplog):
    """Test invalid_response_flags fail_if_any_absent without failing."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "bornhack.dk",
            "family": "ipv4",
            "module": "has_ad",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


# validate_flags fail_if_any_present


def test_validate_flags_fail_if_any_present(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_flags failure by checking for no AD flag on ripe.net."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "ripe.net",
            "family": "ipv4",
            "module": "has_no_ad",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_flags_fail_if_any_present_2(dns_exporter_example_config, caplog):
    """Test invalid_response_flags fail_if_any_present without failing."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "google.com",
            "family": "ipv4",
            "module": "has_no_ad",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


# validate_flags fail_if_all_present


def test_validate_flags_fail_if_all_present(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_flags failure by requiring AA flag from a recursor."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "google.com",
            "family": "ipv4",
            "module": "fail_not_auth",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_flags_fail_if_all_present_2(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_flags failure by expecting an AA+AD response from a recursor."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.quad9.net",
            "query_name": "google.com",
            "family": "ipv4",
            "module": "fail_not_auth",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_flags_fail_if_all_present_3(dns_exporter_example_config, caplog):
    """Test a module using fail_if_all_present without failing by asking an auth and failing on recursive flags."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "k.root-servers.net",
            "query_name": ".",
            "query_type": "NS",
            "family": "ipv4",
            "module": "fail_recursive",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


# validate_flags fail_if_all_absent


def test_validate_flags_fail_if_all_absent(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_flags failure by asking a recursor and expecting no RA flag."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "google.com",
            "family": "ipv4",
            "module": "fail_recursive",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_flags_fail_if_all_absent_2(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_flags failure by not expecting recursive flags from a recursor."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "wikipedia.org",
            "family": "ipv4",
            "module": "fail_recursive",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_flags_fail_if_all_absent_3(dns_exporter_example_config, caplog):
    """Test a module using fail_if_all_absent without failing by asking a root server for . and failing if AA and AD are missing."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "k.root-servers.net",
            "query_name": ".",
            "family": "ipv4",
            "module": "fail_not_auth",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


# validate_rrs fail_if_matches_regexp


def test_validate_rrs_fail_if_matches_regexp(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_answer_rrs failure by asking for root NS and failing on seeing k.root-servers.net. in the result."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "l.root-servers.net",
            "query_name": ".",
            "family": "ipv4",
            "query_type": "NS",
            "module": "fail_auth_k_root",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_rrs_fail_if_matches_regexp_2(dns_exporter_example_config, caplog):
    """Test invalid_response_answer_rrs fail_if_matches_regexp without failing."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv4",
            "query_type": "NS",
            "module": "fail_auth_k_root",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


# validate_rrs fail_if_all_match_regexp


def test_validate_rrs_fail_if_all_match_regexp(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_additional_rrs by asking for root NS and failing on seeing root servers in ADDITIONAL section."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "k.root-servers.net",
            "query_name": ".",
            "family": "ipv4",
            "query_type": "NS",
            "module": "fail_additional_root",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_rrs_fail_if_all_match_regexp_2(dns_exporter_example_config, caplog):
    """Test a module using fail_if_all_match_regexp without failing."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv4",
            "query_type": "NS",
            "module": "fail_additional_root",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


# validate_rrs fail_if_not_matches_regexp


def test_validate_rrs_fail_if_not_matches_regexp(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_answer_rrs by asking for example.com NS and failing on NOT seeing root servers in ANSWER section."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv4",
            "query_type": "NS",
            "module": "fail_answer_root",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_rrs_fail_if_not_matches_regexp_2(dns_exporter_example_config, caplog):
    """Test invalid_response_answer_rrs fail_if_not_matches_regexp without failing."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": ".",
            "family": "ipv4",
            "query_type": "NS",
            "module": "fail_additional_root",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


# validate_rrs fail_if_none_matches_regexp


def test_validate_rrs_fail_if_none_matches_regexp(dns_exporter_example_config, caplog):
    """Trigger an invalid_response_answer_rrs by asking for example.com NS and failing on NOT seeing root servers in ANSWER section."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv4",
            "query_type": "NS",
            "module": "fail_answer_root_none",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_rrs_fail_if_none_matches_regexp_2(
    dns_exporter_example_config,
    caplog,
):
    """Trigger an invalid_response_answer_rrs by asking for root NS and failing on NOT seeing root servers in ANSWER section."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": "example.com",
            "family": "ipv4",
            "query_type": "NS",
            "module": "fail_answer_root_none",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 0.0" in r.text


def test_validate_rrs_fail_if_none_matches_regexp_3(dns_exporter_example_config, caplog):
    """Test fail_if_none_matches with a successful query, make sure k-root is one of the root NS."""
    r = requests.get(
        "http://127.0.0.1:25353/query",
        params={
            "server": "dns.google",
            "query_name": ".",
            "family": "ipv4",
            "query_type": "NS",
            "module": "fail_answer_root_none",
        },
        timeout=5,
    )
    assert "dnsexp_dns_query_success 1.0" in r.text


@pytest.mark.parametrize(
    ("regex_list", "expectation"),
    [
        ([".*"], pytest.raises(ValidationError)),
        ([".*10.8.23.42$"], pytest.raises(ValidationError)),
        ([".*127.0.0.1", ".*CNAME\\sabc.dscc.exampleedge.net"], pytest.raises(ValidationError)),
        ([".*127.0.0.1"], does_not_raise()),
        ([".*127.0.0.1", "192.168.32.42"], does_not_raise()),
        ([".*beef.*"], pytest.raises(ValidationError)),
        (["(?i).*BEEF.*"], pytest.raises(ValidationError)),
        ([".*c0DE.*"], does_not_raise()),
    ],
)
def test_fail_if_matches_regexp(regex_list, expectation, caplog, query):
    """Call DNSCollector.validate_response_rrs() with config for fail_if_matches_regexp.

    consider request failed if any answer rr matches one of these regexes
    """
    test_config = {"name": "test", "validate_answer_rrs": RRValidator.create(fail_if_matches_regexp=regex_list)}

    expected_exception = "Response validator fail_if_matches_regexp failed with reason invalid_response_answer_rrs"

    c = DNSCollector(Config.create(**test_config), query, {})

    caplog.clear()
    caplog.set_level(logging.DEBUG)

    with expectation as e:
        c.validate_response_rrs(test_dns_response)

    assert isinstance(expectation, does_not_raise) or expected_exception in str(e)


@pytest.mark.parametrize(
    ("regex_list", "expectation"),
    [
        ([".*"], pytest.raises(ValidationError)),
        ([".*10.8.23.42", ".*CNAME.*", "(?i).*BEEF.*"], pytest.raises(ValidationError)),
        ([".*10.8.23.42"], does_not_raise()),
        ([".*127.0.0.1"], does_not_raise()),
        ([".*127.0.0.1", ".*CNAME.*"], does_not_raise()),
    ],
)
def test_fail_if_all_match_regexp(regex_list, expectation, caplog, query):
    """Call DNSCollector.validate_response_rrs() with config for fail_if_all_match_regexp.

    consider request failed if all answer rrs match one of these regexes
    """
    test_config = {"name": "test", "validate_answer_rrs": RRValidator.create(fail_if_all_match_regexp=regex_list)}

    expected_exception = "Response validator fail_if_all_match_regexp failed with reason invalid_response_answer_rrs"

    c = DNSCollector(Config.create(**test_config), query, {})

    caplog.clear()
    caplog.set_level(logging.DEBUG)

    with expectation as e:
        c.validate_response_rrs(test_dns_response)

    assert isinstance(expectation, does_not_raise) or expected_exception in str(e)


@pytest.mark.parametrize(
    ("regex_list", "expectation"),
    [
        ([".*"], does_not_raise()),
        ([".*10.8.23.42"], pytest.raises(ValidationError)),
        ([".*10.8.23.42", ".*CNAME.*", ".*beef.*", ".*BEEF.*"], does_not_raise()),
        ([".*10.8.23.42", ".*127.0.0.1"], pytest.raises(ValidationError)),
        ([".*127.0.0.1"], pytest.raises(ValidationError)),
    ],
)
def test_fail_if_not_matches_regexp(regex_list, expectation, caplog, query):
    """Call DNSCollector.validate_response_rrs() with config for fail_if_not_matches_regexp.

    consider request failed if any answer rr does not match one of these regexes
    """
    test_config = {"name": "test", "validate_answer_rrs": RRValidator.create(fail_if_not_matches_regexp=regex_list)}

    expected_exception = "Response validator fail_if_not_matches_regexp failed with reason invalid_response_answer_rrs"

    c = DNSCollector(Config.create(**test_config), query, {})

    caplog.clear()
    caplog.set_level(logging.DEBUG)

    with expectation as e:
        c.validate_response_rrs(test_dns_response)

    assert isinstance(expectation, does_not_raise) or expected_exception in str(e)


@pytest.mark.parametrize(
    ("regex_list", "expectation"),
    [
        ([".*"], does_not_raise()),
        ([".*10.8.23.42"], does_not_raise()),
        ([".*10.8.23.42", ".*CNAME.*"], does_not_raise()),
        ([".*10.8.23.42", ".*127.0.0.1"], does_not_raise()),
        ([".*127.0.0.1"], pytest.raises(ValidationError)),
        ([".*127.0.0.1", "(?i).*ABC"], does_not_raise()),
        ([".*127.0.0.1", ".*ABC"], pytest.raises(ValidationError)),
        ([".*127.0.0.1", ".*XYZ"], pytest.raises(ValidationError)),
    ],
)
def test_fail_if_none_matches_regexp(regex_list, expectation, caplog, query):
    """Call DNSCollector.validate_response_rrs() with config for fail_if_none_matches_regexp.

    consider request failed if none of the answer rrs match one of these regexes
    """
    test_config = {"name": "test", "validate_answer_rrs": RRValidator.create(fail_if_none_matches_regexp=regex_list)}

    expected_exception = "Response validator fail_if_none_matches_regexp failed with reason invalid_response_answer_rrs"

    c = DNSCollector(Config.create(**test_config), query, {})

    caplog.clear()
    caplog.set_level(logging.DEBUG)

    with expectation as e:
        c.validate_response_rrs(test_dns_response)

    assert isinstance(expectation, does_not_raise) or expected_exception in str(e)
